import os
import unittest
from unittest import mock

from kinto.core import events as kinto_events
from pyramid.config import Configurator

from kinto_remote_settings.signer import events as signer_events

from .support import BaseWebTest

here = os.path.abspath(os.path.dirname(__file__))


class Listener(object):
    def __init__(self):
        self.received = []

    def __call__(self, event):
        self.received.append(event)


listener = Listener()


def load_from_config(config, prefix):
    return listener


class ResourceEventsTest(BaseWebTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)

        cls.source_collection = "/buckets/alice/collections/scid"
        cls.destination_collection = "/buckets/destination/collections/dcid"

        settings["kinto.signer.resources"] = "%s -> %s" % (
            cls.source_collection,
            cls.destination_collection,
        )

        settings["kinto.signer.signer_backend"] = (
            "kinto_remote_settings.signer.backends." "local_ecdsa"
        )
        settings["signer.ecdsa.private_key"] = os.path.join(
            here, "config", "ecdsa.private.pem"
        )

        settings["event_listeners"] = "ks"
        settings[
            "event_listeners.ks.use"
        ] = "kinto_remote_settings.signer.tests.test_events"
        return settings

    def setUp(self):
        super().setUp()
        self.app.put_json("/buckets/alice", headers=self.headers)

        resp = self.app.get("/", headers=self.headers)
        self.userid = resp.json["user"]["id"]
        self.app.put_json(
            "/buckets/alice/groups/reviewers",
            {"data": {"members": [self.userid]}},
            headers=self.headers,
        )

        self.app.put_json(self.source_collection, headers=self.headers)
        self.app.post_json(
            self.source_collection + "/records",
            {"data": {"title": "hello"}},
            headers=self.headers,
        )
        self.app.post_json(
            self.source_collection + "/records",
            {"data": {"title": "bonjour"}},
            headers=self.headers,
        )

    def _sign(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.headers,
        )

        resp = self.app.get(self.source_collection, headers=self.headers)
        data = resp.json["data"]
        self.assertEqual(data["status"], "signed")

        resp = self.app.get(self.destination_collection, headers=self.headers)
        data = resp.json["data"]
        self.assertIn("signature", data)

    def test_resource_changed_is_triggered_for_destination_bucket(self):
        self._sign()
        event = [
            e
            for e in listener.received
            if e.payload["uri"] == "/buckets/destination"
            and e.payload["action"] == "create"
        ][0]
        self.assertEqual(len(event.impacted_objects), 1)

        event = [
            e
            for e in listener.received
            if e.payload["uri"] == self.destination_collection
            and e.payload["action"] == "create"
        ][0]
        self.assertEqual(len(event.impacted_objects), 1)
        self.assertEqual(event.payload["user_id"], "plugin:kinto-signer")

    def test_resource_changed_is_triggered_for_work_in_progress(self):
        events = [
            e
            for e in listener.received
            if e.payload["resource_name"] == "collection"
            and e.payload["collection_id"] == "scid"
            and e.payload["action"] == "update"
        ]

        # The first event is when the signer updates the source to mark it as signed.
        self.assertIsNone(events[0].impacted_objects[0]["old"].get("status"))
        self.assertIsNone(events[0].impacted_objects[0]["old"].get("last_edit_date"))
        self.assertEqual(events[0].payload["user_id"], "plugin:kinto-signer")
        self.assertEqual(events[0].impacted_objects[0]["new"]["status"], "signed")

        # We created two records, for each of them we updated the ``last_edit_date``
        # field, so we received two events.
        self.assertIn(
            "basicauth:", events[-1].impacted_objects[0]["new"]["last_edit_by"]
        )
        self.assertNotEqual(
            events[-2].impacted_objects[0]["new"]["last_edit_date"],
            events[-1].impacted_objects[0]["new"]["last_edit_date"],
        )

    def test_resource_changed_is_triggered_for_to_review(self):
        before = len(listener.received)

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )

        events = [
            e
            for e in listener.received[before:]
            if e.payload["resource_name"] == "collection"
            and e.payload["collection_id"] == "scid"
            and e.payload["action"] == "update"
        ]

        self.assertEqual(len(events), 2)
        self.assertIn("basicauth:", events[0].payload["user_id"])
        self.assertEqual(events[0].impacted_objects[0]["new"]["status"], "to-review")
        self.assertNotIn("last_editor", events[0].impacted_objects[0]["new"])

        self.assertEqual(events[1].payload["user_id"], "plugin:kinto-signer")
        self.assertIn(
            "basicauth:", events[1].impacted_objects[0]["new"]["last_review_request_by"]
        )
        self.assertEqual(events[1].impacted_objects[0]["old"]["status"], "to-review")

    def test_resource_changed_is_triggered_for_source_collection(self):
        before = len(listener.received)

        self._sign()
        events = [
            e
            for e in listener.received[before:]
            if e.payload["resource_name"] == "collection"
            and e.payload["collection_id"] == "scid"
            and e.payload["action"] == "update"
        ]
        self.assertEqual(len(events), 2)
        event_tosign = events[0]
        self.assertEqual(len(event_tosign.impacted_objects), 1)
        self.assertEqual(event_tosign.impacted_objects[0]["new"]["status"], "to-sign")
        event_signed = events[1]
        self.assertEqual(len(event_signed.impacted_objects), 1)
        self.assertEqual(event_signed.impacted_objects[0]["old"]["status"], "to-sign")
        self.assertEqual(event_signed.impacted_objects[0]["new"]["status"], "signed")
        self.assertGreater(
            event_signed.payload["timestamp"], event_tosign.payload["timestamp"]
        )

    def test_resource_changed_is_triggered_for_resign(self):
        self._sign()
        before = len(listener.received)

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-resign"}},
            headers=self.headers,
        )

        events = [
            e
            for e in listener.received[before:]
            if e.payload["collection_id"] == "scid"
        ]
        assert len(events) == 2
        event_tosign = events[0]
        assert len(event_tosign.impacted_records) == 1
        assert event_tosign.impacted_records[0]["new"]["status"] == "to-resign"
        event_signed = events[1]
        assert len(event_signed.impacted_records) == 1
        assert event_signed.impacted_records[0]["old"]["status"] == "to-resign"
        assert event_signed.impacted_records[0]["new"]["status"] == "signed"

    def test_resource_changed_is_triggered_for_destination_collection(self):
        before = len(listener.received)

        self._sign()
        event = [
            e
            for e in listener.received[before:]
            if e.payload["resource_name"] == "collection"
            and e.payload.get("collection_id") == "dcid"
            and e.payload["action"] == "update"
        ][0]

        self.assertEqual(len(event.impacted_objects), 1)
        self.assertNotEqual(
            event.impacted_objects[0]["old"].get("signature"),
            event.impacted_objects[0]["new"]["signature"],
        )

    def test_resource_changed_is_triggered_for_destination_creation(self):
        before = len(listener.received)

        self._sign()
        events = [
            e
            for e in listener.received[before:]
            if e.payload["resource_name"] == "record"
            and e.payload["collection_id"] == "dcid"
        ]

        self.assertEqual(len(events), 1)
        self.assertEqual(events[0].payload["action"], "create")
        self.assertEqual(len(events[0].impacted_objects), 2)
        updated = events[0].impacted_objects[0]
        self.assertNotIn("old", updated)
        self.assertIn(updated["new"]["title"], ("bonjour", "hello"))

    def test_resource_changed_is_triggered_for_destination_update(self):
        record_uri = self.source_collection + "/records/xyz"
        self.app.put_json(
            record_uri, {"data": {"title": "salam"}}, headers=self.headers
        )
        self._sign()
        self.app.patch_json(
            record_uri, {"data": {"title": "servus"}}, headers=self.headers
        )

        before = len(listener.received)

        self._sign()
        events = [
            e
            for e in listener.received[before:]
            if e.payload["resource_name"] == "record"
            and e.payload["collection_id"] == "dcid"
        ]

        self.assertEqual(len(events), 1)
        self.assertEqual(events[0].payload["action"], "update")
        self.assertEqual(len(events[0].impacted_objects), 1)
        updated = events[0].impacted_objects[0]
        self.assertIn(updated["old"]["title"], "salam")
        self.assertIn(updated["new"]["title"], "servus")

    def test_resource_changed_is_triggered_for_destination_removal(self):
        record_uri = self.source_collection + "/records/xyz"
        self.app.put_json(
            record_uri, {"data": {"title": "servus"}}, headers=self.headers
        )
        self._sign()
        self.app.delete(record_uri, headers=self.headers)

        before = len(listener.received)
        self._sign()

        events = [
            e
            for e in listener.received[before:]
            if e.payload["resource_name"] == "record"
        ]

        self.assertEqual(len(events), 1)
        self.assertEqual(events[0].payload["action"], "delete")
        self.assertEqual(
            events[0].payload["uri"], self.destination_collection + "/records/xyz"
        )

    def test_resource_changed_is_sent_on_rollback(self):
        # Publish records from setup.
        self._sign()

        records = self.app.get(
            self.source_collection + "/records", headers=self.headers
        ).json["data"]
        self.app.delete(
            self.source_collection + "/records/" + records[0]["id"],
            headers=self.headers,
        )
        self.app.patch_json(
            self.source_collection + "/records/" + records[1]["id"],
            {"data": {"language": "unknown"}},
            headers=self.headers,
        )
        self.app.post_json(
            self.source_collection + "/records",
            {"data": {"title": "some extra record"}},
            headers=self.headers,
        )

        before = len(listener.received)
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-rollback"}},
            headers=self.headers,
        )
        events = [
            e
            for e in listener.received[before:]
            if e.payload["resource_name"] == "record"
        ]

        self.assertEqual(len(events), 3)

        self.assertEqual(events[0].payload["action"], "update")
        self.assertEqual(len(events[0].impacted_objects), 1)
        self.assertEqual(events[0].impacted_objects[0]["new"]["title"], "hello")
        self.assertEqual(events[0].impacted_objects[0]["old"]["language"], "unknown")

        self.assertEqual(events[1].payload["action"], "create")
        self.assertEqual(len(events[1].impacted_objects), 1)
        self.assertEqual(events[1].impacted_objects[0]["new"]["title"], "bonjour")
        self.assertEqual(events[1].impacted_objects[0].get("old"), None)

        self.assertEqual(events[2].payload["action"], "delete")
        self.assertEqual(len(events[2].impacted_objects), 1)
        self.assertEqual(
            events[2].impacted_objects[0]["old"]["title"], "some extra record"
        )
        self.assertEqual(events[2].impacted_objects[0]["new"]["deleted"], True)


class SignoffEventsTest(BaseWebTest, unittest.TestCase):
    events = []

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)

        cls.source_collection = "/buckets/alice/collections/scid"
        cls.destination_collection = "/buckets/destination/collections/dcid"

        settings["kinto.signer.resources"] = "%s -> %s" % (
            cls.source_collection,
            cls.destination_collection,
        )

        settings["kinto.signer.signer_backend"] = (
            "kinto_remote_settings.signer.backends." "local_ecdsa"
        )
        settings["signer.ecdsa.private_key"] = os.path.join(
            here, "config", "ecdsa.private.pem"
        )
        return settings

    @classmethod
    def make_app(cls, settings=None, config=None):
        config = Configurator(settings=cls.get_app_settings())

        def on_review_received(event):
            event.request.registry.storage.create(
                resource_name="custom", parent_id="", record={"pi": 3.14}
            )

        def on_signer_event(event):
            cls.events.append(event)

        config.add_subscriber(on_review_received, signer_events.ReviewRequested)
        config.add_subscriber(on_signer_event, signer_events.ReviewRequested)
        config.add_subscriber(on_signer_event, signer_events.ReviewRejected)
        config.add_subscriber(on_signer_event, signer_events.ReviewApproved)
        config.add_subscriber(on_signer_event, signer_events.ReviewCanceled)

        return super().make_app(config=config)

    def setUp(self):
        del self.events[:]

        self.app.put_json("/buckets/alice", headers=self.headers)

        resp = self.app.get("/", headers=self.headers)
        self.userid = resp.json["user"]["id"]
        self.app.put_json(
            "/buckets/alice/groups/reviewers",
            {"data": {"members": [self.userid]}},
            headers=self.headers,
        )

        self.app.put_json(self.source_collection, headers=self.headers)
        self.app.post_json(
            self.source_collection + "/records",
            {"data": {"title": "hello"}},
            headers=self.headers,
        )
        self.app.post_json(
            self.source_collection + "/records",
            {"data": {"title": "bonjour"}},
            headers=self.headers,
        )

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review", "last_editor_comment": "Double check"}},
            headers=self.headers,
        )

    def test_review_requested_is_triggered(self):
        assert isinstance(self.events[-1], signer_events.ReviewRequested)
        assert self.events[-1].comment == "Double check"

    def test_events_have_details_attributes(self):
        e = self.events[-1]
        assert e.request.path == "/" + self.api_prefix + self.source_collection
        assert e.payload["uri"] == self.source_collection
        assert e.payload["collection_id"] == "scid"
        assert e.impacted_objects[0]["new"]["id"] == "scid"
        assert e.resource["source"]["bucket"] == "alice"
        assert isinstance(e.original_event, kinto_events.ResourceChanged)

    def test_events_have_details_deprecated_attributes(self):
        e = self.events[-1]
        assert e.impacted_records == e.impacted_objects

    def test_review_rejected_is_triggered(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "work-in-progress", "last_reviewer_comment": "Wrong"}},
            headers=self.headers,
        )
        assert isinstance(self.events[-1], signer_events.ReviewRejected)
        assert self.events[-1].comment == "Wrong"

    def test_review_cancelled_is_triggered(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-rollback"}},
            headers=self.headers,
        )
        assert isinstance(self.events[-1], signer_events.ReviewCanceled)

    def test_review_cancelled_is_not_triggered_if_no_change(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-rollback"}},
            headers=self.headers,
        )
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "work-in-progress"}},
            headers=self.headers,
        )
        del self.events[:]
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-rollback"}},
            headers=self.headers,
        )
        assert len(self.events) == 0

    def test_review_requested_is_not_triggered_if_already_set(self):
        r = self.app.get(self.source_collection, headers=self.headers)
        assert r.json["data"]["status"] == "to-review"

        del self.events[:]
        self.app.patch_json(
            self.source_collection, {"data": {"foo": "baz"}}, headers=self.headers
        )
        assert len(self.events) == 0

    def test_review_rejected_is_not_triggered_if_not_waiting_review(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.headers,
        )
        del self.events[:]
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "work-in-progress"}},
            headers=self.headers,
        )
        assert len(self.events) == 0

    def test_review_rejected_is_not_triggered_when_modified_indirectly(self):
        del self.events[:]
        self.app.post_json(
            self.source_collection + "/records",
            {"data": {"title": "hello"}},
            headers=self.headers,
        )
        assert len(self.events) == 0

    def test_review_approved_is_triggered(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.headers,
        )
        assert isinstance(self.events[-1], signer_events.ReviewApproved)
        assert self.events[-1].changes_count == 2

    def test_review_approved_is_not_triggered_on_resign(self):
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.headers,
        )
        del self.events[:]
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-resign"}},
            headers=self.headers,
        )
        assert len(self.events) == 0

    def test_event_is_not_sent_if_rolledback(self):
        patch = mock.patch(
            "kinto_remote_settings.signer.backends.local_ecdsa.ECDSASigner.sign",
            side_effect=ValueError("boom"),
        )
        self.addCleanup(patch.stop)
        patch.start()

        del self.events[:]
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.headers,
            status=500,
        )
        assert len(self.events) == 0

    def test_database_changes_in_subscribers_are_committed(self):
        count = self.storage.count_all(resource_name="custom", parent_id="")
        assert count == 1



import base64
import hashlib
import warnings

import ecdsa
from ecdsa import NIST384p, SigningKey, VerifyingKey

from ..utils import get_first_matching_setting
from .base import SignerBase
from .exceptions import BadSignatureError

# Autograph uses this prefix prior to signing.
SIGN_PREFIX = b"Content-Signature:\x00"


class ECDSASigner(SignerBase):
    def __init__(self, private_key=None, public_key=None):
        if private_key is None and public_key is None:
            msg = "Please, specify either a private_key or public_key " "location."
            raise ValueError(msg)
        self.private_key = private_key
        self.public_key = public_key

    @classmethod
    def generate_keypair(cls):
        sk = SigningKey.generate(curve=NIST384p)
        vk = sk.get_verifying_key()
        return sk.to_pem(), vk.to_pem()

    def load_private_key(self):
        if self.private_key is None:
            msg = "Please, specify the private_key location."
            raise ValueError(msg)

        with open(self.private_key, "rb") as key_file:
            return SigningKey.from_pem(key_file.read())

    def load_public_key(self):
        # Check settings validity
        if self.private_key:
            private_key = self.load_private_key()
            return private_key.get_verifying_key()
        elif self.public_key:
            with open(self.public_key, "rb") as key_file:
                return VerifyingKey.from_pem(key_file.read())

    def sign(self, payload):
        if isinstance(payload, str):  # pragma: nocover
            payload = payload.encode("utf-8")

        payload = SIGN_PREFIX + payload
        private_key = self.load_private_key()
        signature = private_key.sign(
            payload, hashfunc=hashlib.sha384, sigencode=ecdsa.util.sigencode_string
        )
        x5u = ""
        enc_signature = base64.urlsafe_b64encode(signature).decode("utf-8")
        return {"signature": enc_signature, "x5u": x5u, "mode": "p384ecdsa"}

    def verify(self, payload, signature_bundle):
        if isinstance(payload, str):  # pragma: nocover
            payload = payload.encode("utf-8")

        payload = SIGN_PREFIX + payload
        signature = signature_bundle["signature"]
        if isinstance(signature, str):  # pragma: nocover
            signature = signature.encode("utf-8")

        signature_bytes = base64.urlsafe_b64decode(signature)

        public_key = self.load_public_key()
        try:
            public_key.verify(
                signature_bytes,
                payload,
                hashfunc=hashlib.sha384,
                sigdecode=ecdsa.util.sigdecode_string,
            )
        except Exception as e:
            raise BadSignatureError(e)


def load_from_settings(settings, prefix="", *, prefixes=None):
    if prefixes is None:
        prefixes = [prefix]

    if prefix != "":
        message = (
            "signer.load_from_settings `prefix` parameter is deprecated, please "
            "use `prefixes` instead."
        )
        warnings.warn(message, DeprecationWarning)

    private_key = get_first_matching_setting("ecdsa.private_key", settings, prefixes)
    public_key = get_first_matching_setting("ecdsa.public_key", settings, prefixes)
    try:
        return ECDSASigner(private_key=private_key, public_key=public_key)
    except ValueError:
        msg = (
            "Please specify either kinto.signer.ecdsa.private_key or "
            "kinto.signer.ecdsa.public_key in the settings."
        )
        raise ValueError(msg)



import copy

from kinto.core import errors
from kinto.core.errors import ERRORS
from kinto.core.events import ACTIONS
from kinto.core.storage.exceptions import ObjectNotFoundError
from kinto.core.utils import instance_uri
from pyramid import httpexceptions
from pyramid.interfaces import IAuthorizationPolicy

from . import events as signer_events
from .updater import TRACKING_FIELDS, LocalUpdater
from .utils import PLUGIN_USERID, STATUS, ensure_resource_exists

REVIEW_SETTINGS = (
    "reviewers_group",
    "editors_group",
    "to_review_enabled",
)


def raise_invalid(**kwargs):
    kwargs.update(errno=ERRORS.INVALID_POSTED_DATA)
    raise errors.http_error(httpexceptions.HTTPBadRequest(), **kwargs)


def raise_forbidden(**kwargs):
    kwargs.update(errno=ERRORS.FORBIDDEN)
    raise errors.http_error(httpexceptions.HTTPForbidden(), **kwargs)


def pick_resource_and_signer(request, resources, bucket_id, collection_id):
    bucket_key = instance_uri(request, "bucket", id=bucket_id)
    collection_key = instance_uri(
        request, "collection", bucket_id=bucket_id, id=collection_id
    )

    resource = signer = None

    # Review might have been configured explictly for this collection,
    if collection_key in resources:
        resource = resources[collection_key]
    elif bucket_key in resources:
        # Or via its bucket.
        resource = copy.deepcopy(resources[bucket_key])
        # Since it was configured per bucket, we want to make this
        # resource look as if it was configured explicitly for this
        # collection.
        resource["source"]["collection"] = collection_id
        resource["destination"]["collection"] = collection_id
        if "preview" in resource:
            resource["preview"]["collection"] = collection_id

    if collection_key in request.registry.signers:
        signer = request.registry.signers[collection_key]
    elif bucket_key in request.registry.signers:
        signer = request.registry.signers[bucket_key]

    return resource, signer


def resource_group(resource, name, default):
    group = resource.get(name, default)
    # If review is configured per-bucket, the group patterns have to be replaced
    # with the source collection id.
    return group.format(collection_id=resource["source"]["collection"])


def sign_collection_data(event, resources, **kwargs):
    """
    Listen to resource change events, to check if a new signature is
    requested.

    When a source collection specified in settings is modified, and its
    new metadata ``status`` is set to ``"to-sign"``, then sign the data
    and update the destination.
    """
    payload = event.payload

    is_new_collection = payload["action"] == ACTIONS.CREATE.value

    current_user_id = event.request.prefixed_userid
    if current_user_id == PLUGIN_USERID:
        # Ignore changes made by plugin.
        return

    # Prevent recursivity, since the following operations will alter the current
    # collection.
    impacted_objects = list(event.impacted_objects)

    for impacted in impacted_objects:
        new_collection = impacted["new"]
        old_collection = impacted.get("old", {})

        # Only sign the configured resources.
        resource, signer = pick_resource_and_signer(
            event.request,
            resources,
            bucket_id=payload["bucket_id"],
            collection_id=new_collection["id"],
        )
        if resource is None:
            continue

        updater = LocalUpdater(
            signer=signer,
            storage=event.request.registry.storage,
            permission=event.request.registry.permission,
            source=resource["source"],
            destination=resource["destination"],
        )

        uri = instance_uri(
            event.request,
            "collection",
            bucket_id=payload["bucket_id"],
            id=new_collection["id"],
        )

        has_preview_collection = "preview" in resource

        payload = payload.copy()
        payload["uri"] = uri
        payload["collection_id"] = new_collection["id"]

        review_event_cls = None
        review_event_kw = dict(
            request=event.request,
            payload=payload,
            impacted_objects=[impacted],
            resource=resource,
            original_event=event,
        )

        new_status = new_collection.get("status")
        old_status = old_collection.get("status")

        # Autorize kinto-attachment metadata write access. #190
        event.request._attachment_auto_save = True

        if is_new_collection:
            if has_preview_collection:
                updater.destination = resource["preview"]
                updater.sign_and_update_destination(
                    event.request,
                    source_attributes=new_collection,
                    # Do not update source attributes (done below).
                    next_source_status=None,
                )
            updater.destination = resource["destination"]
            updater.sign_and_update_destination(
                event.request,
                source_attributes=new_collection,
                # Prevents last_review_date to be set.
                previous_source_status=STATUS.SIGNED,
                # Signed by default.
                next_source_status=STATUS.SIGNED,
            )

        elif old_status == new_status:
            continue

        elif new_status == STATUS.TO_SIGN:
            # Run signature process (will set `last_reviewer` field).
            if has_preview_collection:
                updater.destination = resource["preview"]
                updater.sign_and_update_destination(
                    event.request,
                    source_attributes=new_collection,
                    previous_source_status=old_status,
                )

            updater.destination = resource["destination"]
            review_event_cls = signer_events.ReviewApproved
            changes_count = updater.sign_and_update_destination(
                event.request,
                source_attributes=new_collection,
                previous_source_status=old_status,
            )
            review_event_kw["changes_count"] = changes_count

        elif new_status == STATUS.TO_REVIEW:
            if has_preview_collection:
                # If preview collection: update and sign preview collection
                updater.destination = resource["preview"]
                changes_count = updater.sign_and_update_destination(
                    event.request,
                    source_attributes=new_collection,
                    next_source_status=STATUS.TO_REVIEW,
                )
            else:
                # If no preview collection: just track `last_editor`
                updater.update_source_review_request_by(event.request)
                changes_count = None
            review_event_cls = signer_events.ReviewRequested
            review_event_kw["changes_count"] = changes_count
            review_event_kw["comment"] = new_collection.get("last_editor_comment", "")

        elif old_status == STATUS.TO_REVIEW and new_status == STATUS.WORK_IN_PROGRESS:
            review_event_cls = signer_events.ReviewRejected
            review_event_kw["comment"] = new_collection.get("last_reviewer_comment", "")

        elif new_status == STATUS.TO_REFRESH:
            updater.refresh_signature(event.request, next_source_status=old_status)
            if has_preview_collection:
                updater.destination = resource["preview"]
                updater.refresh_signature(event.request, next_source_status=old_status)

        elif new_status == STATUS.TO_ROLLBACK:
            # Reset source with destination content, and set status to SIGNED.
            changes_count = updater.rollback_changes(event.request)
            if has_preview_collection:
                # Reset preview with destination content.
                updater.source = resource["preview"]
                changes_count += updater.rollback_changes(
                    event.request, refresh_last_edit=False
                )
                # Refresh signature for this new preview collection content.
                updater.destination = resource["preview"]
                # Without refreshing the source attributes.
                updater.refresh_signature(event.request, next_source_status=None)
            # If some changes were effectively rolledback, send an event.
            if changes_count > 0:
                review_event_cls = signer_events.ReviewCanceled
                review_event_kw["changes_count"] = changes_count

        # Notify request of review.
        if review_event_cls:
            review_event = review_event_cls(**review_event_kw)
            event.request.bound_data.setdefault(
                "kinto_remote_settings.signer.events", []
            ).append(review_event)


def send_signer_events(event):
    """Send accumulated review events for this request. This listener is bound to the
    ``AfterResourceChanged`` event so that review events are sent only if the
    transaction was committed.
    """
    review_events = event.request.bound_data.pop(
        "kinto_remote_settings.signer.events", []
    )
    for review_event in review_events:
        event.request.registry.notify(review_event)


def check_collection_status(
    event,
    resources,
    to_review_enabled,
    editors_group,
    reviewers_group,
):
    """Make sure status changes are allowed."""
    payload = event.payload

    current_user_id = event.request.prefixed_userid
    if current_user_id == PLUGIN_USERID:
        # Ignore changes made by plugin.
        return

    user_principals = event.request.effective_principals

    for impacted in event.impacted_objects:
        old_collection = impacted.get("old", {})
        old_status = old_collection.get("status")
        new_collection = impacted["new"]
        new_status = new_collection.get("status")

        # Skip if collection is not configured for review.
        resource, _ = pick_resource_and_signer(
            event.request,
            resources,
            bucket_id=payload["bucket_id"],
            collection_id=new_collection["id"],
        )
        if resource is None:
            continue

        # to-review and group checking.
        _to_review_enabled = resource.get("to_review_enabled", to_review_enabled)
        _editors_group = resource_group(
            resource, "editors_group", default=editors_group
        )
        _reviewers_group = resource_group(
            resource, "reviewers_group", default=reviewers_group
        )
        # Member of groups have their URIs in their principals.
        editors_group_uri = instance_uri(
            event.request, "group", bucket_id=payload["bucket_id"], id=_editors_group
        )
        reviewers_group_uri = instance_uri(
            event.request, "group", bucket_id=payload["bucket_id"], id=_reviewers_group
        )

        if old_status == new_status:
            # When collection is created old_status == new_status == None.
            continue

        # 0. Nobody can remove the status
        if new_status is None:
            raise_invalid(message="Cannot remove status")

        # 1. None -> work-in-progress
        elif new_status == STATUS.WORK_IN_PROGRESS:
            pass

        # 2. work-in-progress -> to-review
        elif new_status == STATUS.TO_REVIEW:
            if editors_group_uri not in user_principals and _to_review_enabled:
                raise_forbidden(message="Not in %s group" % _editors_group)

        # 3. to-review -> work-in-progress
        # 3. to-review -> to-sign
        # 3. signed -> to-sign
        elif new_status == STATUS.TO_SIGN:
            # Refresh signature (signed -> to-sign) does not require group membership
            if old_status == STATUS.SIGNED:
                raise_invalid(message="Collection already signed")

            if _to_review_enabled:
                # Only allow to-sign from to-review if reviewer and no-editor
                if reviewers_group_uri not in user_principals:
                    raise_forbidden(message="Not in %s group" % _reviewers_group)

                if old_status != STATUS.TO_REVIEW:
                    raise_invalid(message="Collection not under review")

                field_last_requester = TRACKING_FIELDS.LAST_REVIEW_REQUEST_BY.value
                is_same_editor = (
                    old_collection.get(field_last_requester) == current_user_id
                )
                if is_same_editor:
                    raise_forbidden(message="Last editor cannot review")

        # 4. to-sign -> signed
        elif new_status == STATUS.SIGNED:
            raise_invalid(message="Cannot set status to '%s'" % new_status)

        # 5. Refresh signature
        elif new_status == STATUS.TO_REFRESH:
            # Before here we would raise a 400 if the collection had never been
            # signed, but after some thought it does not really make sense.
            pass

        # Rollback changes
        elif new_status == STATUS.TO_ROLLBACK:
            if old_status == STATUS.SIGNED:
                raise_invalid(message="Collection has no work-in-progress")

        # Unknown manual status
        else:
            raise_invalid(message="Invalid status '%s'" % new_status)


def signer_resource_match(resource, bid, cid):
    return resource["bucket"] == bid and (
        resource["collection"] is None or resource["collection"] == cid
    )


def signer_impacts_resource(signer, bid, cid):
    matches_destination = signer_resource_match(signer["destination"], bid, cid)
    if matches_destination:
        return True

    if "preview" in signer:
        matches_preview = signer_resource_match(signer["preview"], bid, cid)
        if matches_preview:
            return True

    return False


def prevent_float_value(event, resources):
    """This ResourceChanged event listener will reject records that
    contain float values.

    In order to be able to align our Canonical JSON implementation with
    the most elaborated specification [0], we must forbid floats to be
    introduced in the system.

    This is indeed the only area where our implementation differs, and
    instead of supporting complex migration paths, getting rid of floats
    is the simplest approach. Floats can be published as strings if needed.

    [0] https://github.com/gibson042/canonicaljson-spec
    """

    def scan(d, path=""):
        for k, v in d.items():
            path = f"{path}.{k}" if path else k
            if isinstance(v, float):
                raise ValueError(
                    f"'{path}' field contains float value (tip: use integer or string)"
                )
            elif isinstance(v, dict):
                scan(v, path)

    # Only raise in configured resources.
    resource, _ = pick_resource_and_signer(
        event.request,
        resources,
        bucket_id=event.payload["bucket_id"],
        collection_id=event.payload["collection_id"],
    )
    if resource is None:
        return

    # Check each created/updated record in the batch.
    for impacted in event.impacted_objects:
        try:
            scan(impacted["new"])
        except ValueError as e:
            raise_invalid(message=str(e))


def prevent_collection_delete(event, resources):
    request = event.request
    bid = event.payload["bucket_id"]
    for impacted in event.impacted_objects:
        cid = impacted["old"]["id"]

        # Locate any collections that imply usage of this collection.
        # If there's some path s -> p -> d for which this collection
        # corresponds to p or d, we forbid deletion of this collection
        # (it's "in use").
        in_use = None

        # The most obvious path is if there is a signer that mentions
        # this collection explicitly in p or d.
        specific_signers = [
            v
            for v in resources.values()
            if v["source"]["collection"] is not None
            and signer_impacts_resource(v, bid, cid)
        ]

        if specific_signers:
            assert (
                len(specific_signers) == 1
            ), f"Inconsistent signers: multiple signers touch {bid} and {cid}"
            in_use = specific_signers[0]

        if not in_use:
            # We identify bucket-wide signers for which p or d matches
            # this collection -- in this case, editing the collection of
            # the same name in s could trigger writes to p or d.
            bucket_signers = [
                v
                for v in resources.values()
                if v["source"]["collection"] is None
                and signer_impacts_resource(v, bid, cid)
            ]
            if bucket_signers:
                assert (
                    len(bucket_signers) == 1
                ), f"Inconsistent signers: multiple signers touch {bid}"
                in_use = bucket_signers[0]

            if in_use:
                # See if this bucket-wide signer is superseded by any
                # specific-collection signers. A specific-collection
                # signer counts as superseding a bucket-wide signer if
                # the specific collection is in the same bucket as the
                # bucket-wide signer, and the specific-collection
                # signer has the same collection ID as the collection
                # being deleted. In this case, we can ignore the
                # bucket-wide s -> p -> d because the
                # collection-specific signer specifies a different
                # workflow for the collection that we thought to
                # impact this one.
                #
                # Specific-collection signers that point *from* other
                # collections to this one are handled explicitly, above.
                #
                # N.B. We can't use signer_impacts_resource here
                # because we want to detect a signer for a
                # specific source collection, regardless of whether it
                # impacts the collection to be deleted or not. A good
                # example where this comes up is where a
                # specific-collection signer disables preview. We want
                # to find this signer even though the preview
                # collection is no longer being impacted.
                for signer in resources.values():
                    same_bucket = (
                        signer["source"]["bucket"] == in_use["source"]["bucket"]
                    )
                    this_collection = signer["source"]["collection"] == cid
                    if same_bucket and this_collection:
                        # Clear the bucket-wide signer.
                        # This signer either named this collection
                        # explicitly (in which case it was handled
                        # above), or it didn't (in which case the
                        # collection is safe to be deleted).
                        in_use = None
                        break

        if in_use is None:
            # Can delete!
            continue

        source_bucket_uri = instance_uri(
            event.request, "bucket", id=in_use["source"]["bucket"]
        )
        source_collection_id = in_use["source"]["collection"] or cid
        try:
            request.registry.storage.get(
                resource_name="collection",
                parent_id=source_bucket_uri,
                object_id=source_collection_id,
            )
            raise_forbidden(message="Collection is in use.")
        except ObjectNotFoundError:
            # Do not prevent delete of preview/destination if source does not exist.
            pass


def check_collection_tracking(event, resources):
    """Make sure tracking fields are not changed manually/removed."""
    if event.request.prefixed_userid == PLUGIN_USERID:
        return

    for impacted in event.impacted_objects:
        old_collection = impacted.get("old", {})
        new_collection = impacted["new"]

        resource, _ = pick_resource_and_signer(
            event.request,
            resources,
            bucket_id=event.payload["bucket_id"],
            collection_id=new_collection["id"],
        )
        # Skip if resource is not configured.
        if resource is None:
            continue

        for field in TRACKING_FIELDS:
            old = old_collection.get(field.value)
            new = new_collection.get(field.value)
            if old != new:
                raise_invalid(message="Cannot change %r" % field)


def set_work_in_progress_status(event, resources):
    """Put the status in work-in-progress if was signed."""
    resource, signer = pick_resource_and_signer(
        event.request,
        resources,
        bucket_id=event.payload["bucket_id"],
        collection_id=event.payload["collection_id"],
    )
    # Skip if resource is not configured.
    if resource is None:
        return

    updater = LocalUpdater(
        signer=signer,
        storage=event.request.registry.storage,
        permission=event.request.registry.permission,
        source=resource["source"],
        destination=resource["destination"],
    )
    updater.update_source_status(STATUS.WORK_IN_PROGRESS, event.request)


def create_editors_reviewers_groups(event, resources, editors_group, reviewers_group):
    if event.request.prefixed_userid == PLUGIN_USERID:
        return

    bid = event.payload["bucket_id"]
    bucket_uri = instance_uri(event.request, "bucket", id=bid)

    current_user_id = event.request.prefixed_userid
    principals = event.request.prefixed_principals

    authz = event.request.registry.getUtility(IAuthorizationPolicy)

    for impacted in event.impacted_objects:
        new_collection = impacted["new"]

        # Skip if collection is not configured for review.
        resource, _ = pick_resource_and_signer(
            event.request,
            resources,
            bucket_id=event.payload["bucket_id"],
            collection_id=new_collection["id"],
        )
        if resource is None:
            continue

        _editors_group = resource_group(
            resource, "editors_group", default=editors_group
        )
        _reviewers_group = resource_group(
            resource, "reviewers_group", default=reviewers_group
        )

        required_perms = authz.get_bound_permissions(bucket_uri, "group:create")
        permission = event.request.registry.permission
        if not permission.check_permission(principals, required_perms):
            return

        group_perms = {"write": [current_user_id]}
        for group, members in (
            (_editors_group, [current_user_id]),
            (_reviewers_group, []),
        ):
            ensure_resource_exists(
                request=event.request,
                resource_name="group",
                parent_id=bucket_uri,
                obj={"id": group, "members": members},
                permissions=group_perms,
                matchdict={"bucket_id": bid, "id": group},
            )

        # Allow those groups to write to the source collection.
        permission = event.request.registry.permission
        collection_uri = instance_uri(
            event.request,
            "collection",
            bucket_id=bid,
            id=resource["source"]["collection"],
        )
        for group in (_editors_group, _reviewers_group):
            group_principal = instance_uri(
                event.request, "group", bucket_id=bid, id=group
            )
            permission.add_principal_to_ace(collection_uri, "write", group_principal)


def cleanup_preview_destination(event, resources):
    storage = event.request.registry.storage

    for impacted in event.impacted_objects:
        old_collection = impacted["old"]

        resource, signer = pick_resource_and_signer(
            event.request,
            resources,
            bucket_id=event.payload["bucket_id"],
            collection_id=old_collection["id"],
        )
        if resource is None:
            continue

        for k in ("preview", "destination"):
            if k not in resource:  # pragma: nocover
                continue
            bid = resource[k]["bucket"]
            cid = resource[k]["collection"]
            collection_uri = instance_uri(
                event.request, "collection", bucket_id=bid, id=cid
            )
            storage.delete_all(
                resource_name="record", parent_id=collection_uri, with_deleted=True
            )

            updater = LocalUpdater(
                signer=signer,
                storage=storage,
                permission=event.request.registry.permission,
                source=resource["source"],
                destination=resource[k],
            )

            # At this point, the DELETE event was sent for the source collection,
            # but the source records may not have been deleted yet (it happens in an
            # event listener too). That's why we don't copy the records otherwise it
            # will recreate the records that were just deleted.
            updater.sign_and_update_destination(
                event.request,
                source_attributes=old_collection,
                next_source_status=None,
                push_records=False,
            )



import json

from kinto_remote_settings.signer.serializer import canonical_json

#
# Kinto specific
#


def test_supports_records_as_iterators():
    records = iter([{"bar": "baz", "last_modified": "45678", "id": "1"}])
    canonical_json(records, "45678")


def test_provides_records_in_data_along_last_modified():
    records = [{"bar": "baz", "last_modified": "45678", "id": "1"}]
    serialized = json.loads(canonical_json(records, "45678"))
    assert "data" in serialized
    assert "last_modified" in serialized


def test_orders_records_by_id():
    records = [
        {"bar": "baz", "last_modified": "45678", "id": "2"},
        {"foo": "bar", "last_modified": "12345", "id": "1"},
    ]
    serialized = json.loads(canonical_json(records, "45678"))
    assert serialized["last_modified"] == "45678"
    assert serialized["data"][0]["id"] == "1"
    assert serialized["data"][1]["id"] == "2"


def test_removes_deleted_items():
    record = {"bar": "baz", "last_modified": "45678", "id": "2"}
    deleted_record = {"deleted": True, "last_modified": "12345", "id": "1"}
    records = [deleted_record, record]
    serialized = canonical_json(records, "42")
    assert [record] == json.loads(serialized)["data"]


#
# Standard
#


def test_does_not_alter_records():
    records = [
        {"foo": "bar", "last_modified": "12345", "id": "1"},
        {"bar": "baz", "last_modified": "45678", "id": "2"},
    ]
    canonical_json(records, "45678")

    assert records == [
        {"foo": "bar", "last_modified": "12345", "id": "1"},
        {"bar": "baz", "last_modified": "45678", "id": "2"},
    ]


def test_preserves_data():
    records = [
        {"foo": "bar", "last_modified": "12345", "id": "1"},
        {"bar": "baz", "last_modified": "45678", "id": "2"},
    ]
    serialized = canonical_json(records, "45678")
    assert records == json.loads(serialized)["data"]



from datetime import datetime, timedelta
from urllib.parse import urlencode

import colander
import kinto.core
from cornice.validators import colander_validator
from kinto.authorization import RouteFactory
from kinto.core import resource
from kinto.core import utils as core_utils
from kinto.core.storage import Filter, Sort
from kinto.core.storage import exceptions as storage_exceptions
from kinto.core.storage.memory import extract_object_set
from kinto.core.utils import COMPARISON, instance_uri
from pyramid import httpexceptions
from pyramid.security import IAuthorizationPolicy
from zope.interface import implementer

from . import (
    CHANGES_COLLECTION,
    CHANGES_COLLECTION_PATH,
    CHANGES_RECORDS_PATH,
    CHANGESET_PATH,
    MONITOR_BUCKET,
)
from .utils import changes_object, monitored_collections


class ChangesModel(object):
    id_field = "id"
    modified_field = "last_modified"
    deleted_field = "deleted"
    permissions_field = "__permissions__"

    def __init__(self, request):
        self.request = request
        self.storage = request.registry.storage

        self.__entries = None

    def timestamp(self):
        if not self._entries():
            return core_utils.msec_time()
        max_value = max([e["last_modified"] for e in self._entries()])
        return max_value

    def get_objects(
        self,
        filters=None,
        sorting=None,
        pagination_rules=None,
        limit=None,
        include_deleted=False,
        parent_id=None,
    ):
        objs, _ = extract_object_set(
            objects=self._entries(),
            filters=filters,
            sorting=sorting,
            pagination_rules=pagination_rules,
            limit=limit,
        )
        return objs

    def _entries(self):
        if self.__entries is None:
            self.__entries = {}

            for (bucket_id, collection_id) in monitored_collections(
                self.request.registry
            ):
                collection_uri = core_utils.instance_uri(
                    self.request, "collection", bucket_id=bucket_id, id=collection_id
                )
                timestamp = self.storage.resource_timestamp(
                    parent_id=collection_uri, resource_name="record"
                )
                entry = changes_object(
                    self.request, bucket_id, collection_id, timestamp
                )
                self.__entries[entry[self.id_field]] = entry

        return self.__entries.values()


class ChangesSchema(resource.ResourceSchema):
    host = colander.SchemaNode(colander.String())
    bucket = colander.SchemaNode(colander.String())
    collection = colander.SchemaNode(colander.String())

    class Options:
        preserve_unknown = False


@implementer(IAuthorizationPolicy)
class AnonymousRoute(RouteFactory):
    def check_permission(self, principals, bound_perms):
        # Bypass permissions check on /buckets/monitor.
        return True


@resource.register(
    name="changes",
    description="List of changes",
    plural_path=CHANGES_RECORDS_PATH,
    object_path=None,
    plural_methods=("GET",),
    factory=AnonymousRoute,
)
class Changes(resource.Resource):

    schema = ChangesSchema

    def __init__(self, request, context=None):
        # Bypass call to storage if _since is too old.
        _handle_old_since_redirect(request)
        # Inject custom model.
        self.model = ChangesModel(request)
        super(Changes, self).__init__(request, context)

    def plural_get(self):
        result = super().plural_get()
        _handle_cache_expires(self.request, MONITOR_BUCKET, CHANGES_COLLECTION)
        return result


def _handle_cache_expires(request, bid, cid):
    # If the client sends cache busting query parameters, then we can cache more
    # aggressively.
    settings = request.registry.settings
    prefix = f"{bid}.{cid}.record_cache"
    default_expires = settings.get(f"{prefix}_expires_seconds")
    maximum_expires = settings.get(f"{prefix}_maximum_expires_seconds", default_expires)

    has_cache_busting = "_expected" in request.GET
    cache_expires = maximum_expires if has_cache_busting else default_expires

    if cache_expires is not None:
        request.response.cache_expires(seconds=int(cache_expires))


def _handle_old_since_redirect(request):
    """
    In order to limit the number of possible combinations
    of `_since` and `_expected` querystring parameters,
    and thus maximize the effect of caching, we redirect the clients
    that arrive here with a very old `_since` value.

    This simply means that these clients will have to iterate
    and compare the local timestamps of the whole list of changes
    instead of a filtered subset.

    https://searchfox.org/mozilla-central/rev/b58ca450/services/settings/remote-settings.js#299

    See https://bugzilla.mozilla.org/show_bug.cgi?id=1529685
    and https://bugzilla.mozilla.org/show_bug.cgi?id=1665319#c2
    """
    try:
        # request.validated is not populated yet (resource was not instantiated yet,
        # we want to bypass storage).
        qs_since_str = request.GET.get("_since", "")
        qs_since = int(qs_since_str.strip('"'))
    except ValueError:
        # Will fail later during resource querystring validation.
        return

    settings = request.registry.settings
    max_age_since = int(settings.get("changes.since_max_age_days", 21))
    if max_age_since < 0:
        # Redirect is disabled.
        return

    min_since_dt = datetime.now() - timedelta(days=max_age_since)
    min_since = min_since_dt.timestamp() * 1000

    if qs_since >= min_since:
        # Since value is recent. No redirect.
        return

    http_scheme = settings.get("http_scheme") or "https"
    http_host = settings.get(
        "changes.http_host", request.registry.settings.get("http_host")
    )
    host_uri = f"{http_scheme}://{http_host}"
    redirect = host_uri + request.matched_route.generate(request.matchdict)

    queryparams = request.GET.copy()
    del queryparams["_since"]
    if queryparams:
        redirect += "?" + urlencode(queryparams)

    # Serve a redirection, with optional cache control headers.
    response = httpexceptions.HTTPTemporaryRedirect(redirect)
    cache_seconds = int(
        settings.get("changes.since_max_age_redirect_ttl_seconds", 86400)
    )
    if cache_seconds >= 0:
        response.cache_expires(cache_seconds)
    raise response


@implementer(IAuthorizationPolicy)
class ChangeSetRoute(RouteFactory):
    """The changeset endpoint should have the same permissions as the collection
    metadata.

    The permission to read records is implicit when metadata are readable.
    """

    def __init__(self, request):
        super().__init__(request)
        bid = request.matchdict["bid"]
        cid = request.matchdict["cid"]
        collection_uri = instance_uri(request, "collection", bucket_id=bid, id=cid)
        # This route context will be the same as when reaching the collection URI.
        self.permission_object_id = collection_uri
        self.required_permission = "read"

    def check_permission(self, principals, bound_perms):
        # The monitor/changes changeset endpoint is publicly accesible.
        if self.permission_object_id == CHANGES_COLLECTION_PATH:
            return True
        # Otherwise rely on the collection permissions.
        return super().check_permission(principals, bound_perms)


changeset = kinto.core.Service(
    name="collection-changeset", path=CHANGESET_PATH, factory=ChangeSetRoute
)


class QuotedTimestamp(colander.SchemaNode):
    """Integer between "" used in _since querystring."""

    schema_type = colander.String
    error_message = "The value should be integer between double quotes."
    validator = colander.Regex('^"([0-9]+?)"$', msg=error_message)

    def deserialize(self, cstruct=colander.null):
        param = super(QuotedTimestamp, self).deserialize(cstruct)
        if param is colander.drop:
            return param
        return int(param.strip('"'))


class ChangeSetQuerystring(colander.MappingSchema):
    _since = QuotedTimestamp(missing=colander.drop)
    _expected = colander.SchemaNode(colander.String())
    _limit = colander.SchemaNode(colander.Integer(), missing=colander.drop)
    # Query parameters used on monitor/changes endpoint.
    bucket = colander.SchemaNode(colander.String(), missing=colander.drop)
    collection = colander.SchemaNode(colander.String(), missing=colander.drop)


class ChangeSetSchema(colander.MappingSchema):
    querystring = ChangeSetQuerystring()


@changeset.get(
    schema=ChangeSetSchema(), permission="read", validators=(colander_validator,)
)
def get_changeset(request):
    bid = request.matchdict["bid"]
    cid = request.matchdict["cid"]

    storage = request.registry.storage

    queryparams = request.validated["querystring"]
    limit = queryparams.get("_limit")
    filters = []
    include_deleted = False
    if "_since" in queryparams:
        filters = [Filter("last_modified", queryparams["_since"], COMPARISON.GT)]
        # Include tombstones when querying with _since
        include_deleted = True

    if (bid, cid) == (MONITOR_BUCKET, CHANGES_COLLECTION):
        # Redirect old since, on monitor/changes only.
        _handle_old_since_redirect(request)

        if "bucket" in queryparams:
            filters.append(Filter("bucket", queryparams["bucket"], COMPARISON.EQ))

        if "collection" in queryparams:
            filters.append(
                Filter("collection", queryparams["collection"], COMPARISON.EQ)
            )

        model = ChangesModel(request)
        metadata = {}
        timestamp = model.timestamp()
        changes = model.get_objects(
            filters=filters, limit=limit, include_deleted=include_deleted
        )

    else:
        bucket_uri = instance_uri(request, "bucket", id=bid)
        collection_uri = instance_uri(request, "collection", bucket_id=bid, id=cid)

        try:
            # We'll make sure that data isn't changed while we read metadata, changes,
            # etc.
            before = storage.resource_timestamp(
                resource_name="record", parent_id=collection_uri
            )
            # Fetch collection metadata.
            metadata = storage.get(
                resource_name="collection", parent_id=bucket_uri, object_id=cid
            )

        except storage_exceptions.ObjectNotFoundError:
            raise httpexceptions.HTTPNotFound()

        except storage_exceptions.BackendError as e:
            # The call to `resource_timestamp()` on an empty collection will try
            # initialize it. If the instance is read-only, it fails with a backend
            # error. Raise 404 in this case otherwise raise the original backend error.
            if "when running in readonly" in str(e):
                raise httpexceptions.HTTPNotFound()
            raise

        # Fetch list of changes.
        changes = storage.list_all(
            resource_name="record",
            parent_id=collection_uri,
            filters=filters,
            limit=limit,
            id_field="id",
            modified_field="last_modified",
            deleted_field="deleted",
            sorting=[Sort("last_modified", -1)],
            include_deleted=include_deleted,
        )
        # Fetch current collection timestamp.
        timestamp = storage.resource_timestamp(
            resource_name="record", parent_id=collection_uri
        )

        # Do not serve inconsistent data.
        if before != timestamp:  # pragma: no cover
            raise storage_exceptions.IntegrityError(message="Inconsistent data. Retry.")

    # Cache control.
    _handle_cache_expires(request, bid, cid)

    # Set Last-Modified response header (Pyramid takes care of converting).
    request.response.last_modified = timestamp / 1000.0

    data = {
        "metadata": metadata,
        "timestamp": timestamp,
        "changes": changes,
    }
    return data



import configparser
import os

from kinto import main as kinto_main
from kinto.core.testing import BaseWebTest as CoreWebTest
from kinto.core.testing import get_user_headers

here = os.path.abspath(os.path.dirname(__file__))


class BaseWebTest(CoreWebTest):
    api_prefix = "v1"
    entry_point = kinto_main
    config = "config.ini"

    def __init__(self, *args, **kwargs):
        super(BaseWebTest, self).__init__(*args, **kwargs)
        self.headers.update(get_user_headers("mat"))
        self.headers.update({"Origin": "http://localhost:9999"})

    @classmethod
    def get_app_settings(cls, extras=None):
        ini_path = os.path.join(here, cls.config)
        config = configparser.ConfigParser()
        config.read(ini_path)
        settings = dict(config.items("app:main"))
        settings.update(extras or {})
        return settings

    def setUp(self):
        super(BaseWebTest, self).setUp()
        self.create_collection("blocklists", "certificates")

    def create_collection(self, bucket_id, collection_id):
        bucket_uri = "/buckets/%s" % bucket_id
        self.app.put_json(bucket_uri, {}, headers=self.headers)
        collection_uri = bucket_uri + "/collections/%s" % collection_id
        self.app.put_json(collection_uri, {}, headers=self.headers)



import unittest
from email.utils import parsedate_to_datetime
from unittest import mock

from kinto.core.storage import exceptions as storage_exceptions
from kinto.core.testing import get_user_headers

from . import BaseWebTest

SAMPLE_RECORD = {"data": {"dev-edition": True}}


class ChangesetViewTest(BaseWebTest, unittest.TestCase):
    records_uri = "/buckets/blocklists/collections/certificates/records"
    changeset_uri = (
        "/buckets/blocklists/collections/certificates/changeset?_expected=42"
    )

    def setUp(self):
        super(ChangesetViewTest, self).setUp()
        self.app.post_json(self.records_uri, SAMPLE_RECORD, headers=self.headers)

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["blocklists.certificates.record_cache_expires_seconds"] = 1234
        return settings

    def test_changeset_is_accessible(self):
        resp = self.app.head(self.records_uri, headers=self.headers)
        records_timestamp = int(resp.headers["ETag"][1:-1])

        resp = self.app.get(self.changeset_uri, headers=self.headers)
        data = resp.json

        assert "metadata" in data
        assert "timestamp" in data
        assert "changes" in data
        assert data["metadata"]["id"] == "certificates"
        assert len(data["changes"]) == 1
        assert data["changes"][0]["dev-edition"] is True
        assert data["timestamp"] == records_timestamp

    def test_last_modified_header_is_set(self):
        resp = self.app.get(self.changeset_uri, headers=self.headers)
        timestamp = resp.json["timestamp"]

        dt = parsedate_to_datetime(resp.headers["Last-Modified"])

        assert dt.timestamp() == int(timestamp / 1000)

    def test_changeset_can_be_filtered(self):
        resp = self.app.post_json(self.records_uri, {}, headers=self.headers)
        before = resp.json["data"]["last_modified"]
        self.app.post_json(self.records_uri, {}, headers=self.headers)

        resp = self.app.get(self.changeset_uri, headers=self.headers)
        assert len(resp.json["changes"]) == 3

        resp = self.app.get(
            self.changeset_uri + f'&_since="{before}"', headers=self.headers
        )
        assert len(resp.json["changes"]) == 1

    def test_tombstones_are_returned(self):
        resp = self.app.get(self.records_uri, headers=self.headers)
        before = resp.headers["ETag"]
        # Delete one record.
        self.app.delete(self.records_uri + "?_limit=1", headers=self.headers)

        resp = self.app.get(
            self.changeset_uri + f"&_since={before}", headers=self.headers
        )

        assert len(resp.json["changes"]) == 1
        assert "deleted" in resp.json["changes"][0]

    def test_changeset_is_not_publicly_accessible(self):
        # By default other users cannot read.
        user_headers = {
            **self.headers,
            **get_user_headers("some-user"),
        }
        self.app.get(self.changeset_uri, status=401)
        self.app.get(self.changeset_uri, headers=user_headers, status=403)

        # Add read permissions to everyone.
        self.app.patch_json(
            "/buckets/blocklists",
            {"permissions": {"read": ["system.Everyone"]}},
            headers=self.headers,
        )

        self.app.get(self.changeset_uri, headers=user_headers, status=200)
        self.app.get(self.changeset_uri, status=200)

    def test_changeset_returns_404_if_collection_is_unknown(self):
        changeset_uri = "/buckets/blocklists/collections/fjuvrb/changeset?_expected=42"
        self.app.get(changeset_uri, headers=self.headers, status=404)

    def test_timestamp_is_validated(self):
        self.app.get(
            self.changeset_uri + "&_since=abc", headers=self.headers, status=400
        )
        self.app.get(
            self.changeset_uri + "&_since=42", headers=self.headers, status=400
        )
        self.app.get(
            self.changeset_uri + "&_since=*)(!(objectClass=*)",
            headers=self.headers,
            status=400,
        )
        self.app.get(self.changeset_uri + '&_since="42"', headers=self.headers)

    def test_expected_param_is_mandatory(self):
        self.app.get(self.changeset_uri.split("?")[0], headers=self.headers, status=400)

    def test_limit_is_supported(self):
        self.app.post_json(self.records_uri, {}, headers=self.headers)

        resp = self.app.get(self.changeset_uri + "&_limit=1", headers=self.headers)
        assert len(resp.json["changes"]) == 1

    def test_extra_param_is_allowed(self):
        self.app.get(self.changeset_uri + "&_extra=abc", headers=self.headers)

    def test_cache_control_headers_are_set(self):
        resp = self.app.get(self.changeset_uri, headers=self.headers)
        assert resp.headers["Cache-Control"] == "max-age=1234"

    def test_raises_original_backend_errors(self):
        backend = self.app.app.registry.storage
        with mock.patch.object(backend, "resource_timestamp") as mocked:
            mocked.side_effect = storage_exceptions.BackendError
            changeset_uri = (
                "/buckets/blocklists/collections/certificates/changeset?_expected=42"
            )
            self.app.get(changeset_uri, headers=self.headers, status=503)


class ReadonlyTest(BaseWebTest, unittest.TestCase):
    changeset_uri = "/buckets/monitor/collections/changes/changeset?_expected=42"

    def setUp(self):
        super().setUp()
        # Mark storage as readonly.
        # We can't do it from test app settings because we need
        # the initial bucket and collection).
        self.app.app.registry.storage.readonly = True

    def test_changeset_returns_404_if_collection_is_unknown(self):
        changeset_uri = "/buckets/blocklists/collections/fjuvrb/changeset?_expected=42"
        self.app.get(changeset_uri, headers=self.headers, status=404)


class MonitorChangesetViewTest(BaseWebTest, unittest.TestCase):
    records_uri = "/buckets/blocklists/collections/{cid}/records"
    changeset_uri = "/buckets/monitor/collections/changes/changeset?_expected=42"

    def setUp(self):
        super().setUp()
        self.create_collection("blocklists", "cfr")
        self.app.post_json(
            self.records_uri.format(cid="cfr"), SAMPLE_RECORD, headers=self.headers
        )
        self.app.post_json(
            self.records_uri.format(cid="certificates"),
            SAMPLE_RECORD,
            headers=self.headers,
        )

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["kinto.changes.since_max_age_days"] = 1
        return settings

    def test_changeset_exists_for_monitor_changes(self):
        resp = self.app.head(
            self.records_uri.format(cid="certificates"), headers=self.headers
        )
        records_timestamp = int(resp.headers["ETag"].strip('"'))

        resp = self.app.get(self.changeset_uri)
        data = resp.json

        assert data["timestamp"] == records_timestamp
        assert len(data["changes"]) == 2
        assert data["changes"][0]["collection"] == "certificates"

    def test_changeset_redirects_if_since_is_too_old(self):
        resp = self.app.get(self.changeset_uri + '&_since="42"')

        assert resp.status_code == 307
        assert resp.headers["Location"] == (
            "https://www.kinto-storage.org/v1"
            "/buckets/monitor/collections/changes/changeset?_expected=42"
        )

    def test_limit_is_supported(self):
        resp = self.app.get(self.changeset_uri + "&_limit=1", headers=self.headers)
        assert len(resp.json["changes"]) == 1

    def test_filter_by_collection(self):
        resp = self.app.get(
            self.changeset_uri + "&bucket=blocklists&collection=cfr",
            headers=self.headers,
        )
        assert len(resp.json["changes"]) == 1



import random
import re
import string
import unittest
from unittest import mock

from .support import BaseWebTest, get_user_headers

RE_ISO8601 = re.compile(r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{6}\+00:00")


class SignerAttachmentsTest(BaseWebTest, unittest.TestCase):
    def setUp(self):
        super().setUp()
        # Patch calls to Autograph.
        patch = mock.patch("kinto_remote_settings.signer.backends.autograph.requests")
        self.addCleanup(patch.stop)
        self.mocked_autograph = patch.start()

        def fake_sign():
            fake_signature = "".join(random.sample(string.ascii_lowercase, 10))
            return [
                {
                    "signature": "",
                    "hash_algorithm": "",
                    "signature_encoding": "",
                    "content-signature": fake_signature,
                    "x5u": "",
                    "ref": "",
                }
            ]

        self.mocked_autograph.post.return_value.json.side_effect = fake_sign

        self.headers = get_user_headers("tarte:en-pion")
        resp = self.app.get("/", headers=self.headers)
        self.userid = resp.json["user"]["id"]

        self.other_headers = get_user_headers("Sam:Wan Heilss")
        resp = self.app.get("/", headers=self.other_headers)
        self.other_userid = resp.json["user"]["id"]

        self.app.put_json(self.source_bucket, headers=self.headers)
        self.app.put_json(self.source_collection, headers=self.headers)
        self.app.put_json(
            self.source_bucket + "/groups/reviewers",
            {"data": {"members": [self.other_userid]}},
            headers=self.headers,
        )

    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)

        settings["kinto.includes"] += " kinto_attachment"

        cls.source_bucket = "/buckets/alice"
        cls.source_collection = cls.source_bucket + "/collections/scid"
        cls.destination_bucket = "/buckets/alice"
        cls.destination_collection = cls.destination_bucket + "/collections/dcid"

        settings["kinto.signer.resources"] = "%s -> %s" % (
            cls.source_collection,
            cls.destination_collection,
        )

        settings["signer.to_review_enabled"] = "false"

        settings["attachment.base_path"] = "/tmp"

        return settings

    def initialize(self):
        r = self.app.post_json(
            self.source_collection + "/records",
            {"data": {"title": "hello"}},
            headers=self.headers,
        )
        r = r.json["data"]
        uri = self.source_collection + "/records/" + r["id"] + "/attachment"
        self.upload_file(
            uri=uri,
            files=[("attachment", "image.jpg", b"--fake--")],
            headers=self.headers,
        )

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.other_headers,
        )

    def upload_file(self, uri, files, params=[], headers={}):
        content_type, body = self.app.encode_multipart(params, files)
        headers = headers.copy()
        headers["Content-Type"] = content_type
        resp = self.app.post(uri, body, headers=headers)
        return resp

    def test_attachment_is_enabled(self):
        r = self.app.get("/")
        assert "attachments" in r.json["capabilities"]

    def test_attachment_is_published_on_final_collection(self):
        self.initialize()

        r = self.app.get(self.destination_collection + "/records", headers=self.headers)
        record = r.json["data"][0]

        assert "attachment" in record

    def test_attachment_can_be_replaced(self):
        self.initialize()

        r = self.app.get(self.destination_collection + "/records", headers=self.headers)
        record = r.json["data"][0]

        attachment_before = record["attachment"]["hash"]

        uri = self.source_collection + "/records/" + record["id"] + "/attachment"
        self.upload_file(
            uri=uri,
            files=[("attachment", "image.jpg", b"--other-fake--")],
            headers=self.headers,
        )

        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-review"}},
            headers=self.headers,
        )
        self.app.patch_json(
            self.source_collection,
            {"data": {"status": "to-sign"}},
            headers=self.other_headers,
        )

        r = self.app.get(self.destination_collection + "/records", headers=self.headers)
        record = r.json["data"][0]
        attachment_after = record["attachment"]["hash"]

        assert attachment_before != attachment_after



class SignerBase(object):
    def sign(self, payload):
        """
        Signs the specified `payload` and returns the signature metadata.

        :returns: A mapping with every attributes about the signature
            (e.g. "signature", "hash_algorithm", "signature_encoding"...)
        :rtype: dict
        """
        raise NotImplementedError



import copy
import functools
import re

from .. import __version__
from .events import ReviewApproved, ReviewRejected, ReviewRequested

DEFAULT_SIGNER = "kinto_remote_settings.signer.backends.local_ecdsa"


def get_exposed_resources(resource_dict, review_settings):
    """Compute a set of resources to be shown as part of the server's capabilities.

    This should include review settings for each resource but nothing
    related to the actual signing parameters for those resources."""
    out = []
    for resource in resource_dict.values():
        sanitized = {}
        for setting in ["source", "destination", "preview"] + list(review_settings):
            if setting in resource:
                sanitized[setting] = resource[setting]
        out.append(sanitized)

    return out


def on_review_approved(event):
    statsd_client = event.request.registry.statsd
    if statsd_client is not None:
        count = event.changes_count
        bid = event.resource["destination"]["bucket"]
        cid = event.resource["destination"]["collection"]
        # Report into a global counter.
        statsd_client.count("plugins.signer.approved_changes", count)
        # Report for this collection.
        # TODO: when using Datadog, we could annotate the above with tags instead.
        statsd_client.count(f"plugins.signer.approved_changes.{bid}.{cid}", count)


def includeme(config):
    # We import stuff here, so that kinto-signer can be installed with `--no-deps`
    # and used without having this Pyramid ecosystem installed.
    import transaction
    from kinto.core.events import ACTIONS, ResourceChanged
    from pyramid.events import NewRequest
    from pyramid.exceptions import ConfigurationError
    from pyramid.settings import asbool

    from . import listeners, utils
    from .backends import heartbeat

    # Register heartbeat to check signer integration.
    config.registry.heartbeats["signer"] = heartbeat

    settings = config.get_settings()

    # Check source and destination resources are configured.
    raw_resources = settings.get("signer.resources")
    if raw_resources is None:
        error_msg = "Please specify the kinto.signer.resources setting."
        raise ConfigurationError(error_msg)
    resources = utils.parse_resources(raw_resources)

    # Expand the resources with the ones that come from per-bucket resources
    # and have specific settings.
    # For example, consider the case where resource is ``/buckets/dev -> /buckets/prod``
    # and there is a setting ``signer.dev.recipes.signer_backend = foo``
    output_resources = resources.copy()
    for key, resource in resources.items():
        # If collection is not None, there is nothing to expand :)
        if resource["source"]["collection"] is not None:
            continue
        bid = resource["source"]["bucket"]
        # Match setting names like signer.stage.specific.autograph.hawk_id
        matches = [
            (v, re.search(rf"signer\.{bid}\.([^\.]+)\.(.+)", k))
            for k, v in settings.items()
        ]
        found = [(v, m.group(1), m.group(2)) for (v, m) in matches if m]
        # Expand the list of resources with the ones that contain collection
        # specific settings.
        for setting_value, cid, setting_name in found:
            signer_key = f"/buckets/{bid}/collections/{cid}"
            if signer_key not in output_resources:
                specific = copy.deepcopy(resource)
                specific["source"]["collection"] = cid
                specific["destination"]["collection"] = cid
                if "preview" in specific:
                    specific["preview"]["collection"] = cid
                output_resources[signer_key] = specific
            output_resources[signer_key][setting_name] = setting_value
    resources = output_resources

    # Determine which are the settings that apply to all buckets/collections.
    defaults = {
        "reviewers_group": "reviewers",
        "editors_group": "editors",
        "to_review_enabled": False,
    }
    global_settings = {}
    for setting in listeners.REVIEW_SETTINGS:
        value = settings.get("signer.%s" % setting, defaults[setting])
        if setting.endswith("_enabled"):
            value = asbool(value)
        global_settings[setting] = value

    # For each resource that is configured, we determine what signer is
    # configured and what are the review settings.
    # Note: the `resource` values are mutated in place.
    config.registry.signers = {}
    for signer_key, resource in resources.items():
        bid = resource["source"]["bucket"]
        server_wide = "signer."
        bucket_wide = f"signer.{bid}."
        prefixes = [bucket_wide, server_wide]

        per_bucket_config = resource["source"]["collection"] is None

        if not per_bucket_config:
            cid = resource["source"]["collection"]
            collection_wide = f"signer.{bid}.{cid}."
            deprecated = f"signer.{bid}_{cid}."
            prefixes = [collection_wide, deprecated] + prefixes

        # Instantiates the signers associated to this resource.
        dotted_location = utils.get_first_matching_setting(
            "signer_backend", settings, prefixes, default=DEFAULT_SIGNER
        )
        signer_module = config.maybe_dotted(dotted_location)
        backend = signer_module.load_from_settings(settings, prefixes=prefixes)
        config.registry.signers[signer_key] = backend

        # Load the setttings associated to each resource.
        for setting in listeners.REVIEW_SETTINGS:
            # Per collection/bucket:
            value = utils.get_first_matching_setting(
                setting, settings, prefixes, default=global_settings[setting]
            )

            if setting.endswith("_enabled"):
                value = asbool(value)

            # Resolve placeholder with source info.
            if setting.endswith("_group"):
                # If configured per bucket, then we leave the placeholder.
                # It will be resolved in listeners during group checking and
                # by Kinto-Admin when matching user groups with info from capabilities.
                collection_id = resource["source"]["collection"] or "{collection_id}"
                try:
                    value = value.format(
                        bucket_id=resource["source"]["bucket"],
                        collection_id=collection_id,
                    )
                except KeyError as e:
                    raise ConfigurationError("Unknown group placeholder %s" % e)

            # Only expose if relevant.
            if value != global_settings[setting]:
                resource[setting] = value
            else:
                resource.pop(setting, None)

    # Expose the capabilities in the root endpoint.
    exposed_resources = get_exposed_resources(resources, listeners.REVIEW_SETTINGS)
    message = "Digital signatures for integrity and authenticity of records."
    docs = "https://github.com/Kinto/kinto-signer#kinto-signer"
    config.add_api_capability(
        "signer",
        message,
        docs,
        version=__version__,
        resources=exposed_resources,
        # Backward compatibility with < v26
        group_check_enabled=True,
        **global_settings,
    )

    config.add_subscriber(on_review_approved, ReviewApproved)

    config.add_subscriber(
        functools.partial(listeners.set_work_in_progress_status, resources=resources),
        ResourceChanged,
        for_resources=("record",),
    )

    config.add_subscriber(
        functools.partial(
            listeners.check_collection_status, resources=resources, **global_settings
        ),
        ResourceChanged,
        for_actions=(ACTIONS.CREATE, ACTIONS.UPDATE),
        for_resources=("collection",),
    )

    config.add_subscriber(
        functools.partial(listeners.check_collection_tracking, resources=resources),
        ResourceChanged,
        for_actions=(ACTIONS.CREATE, ACTIONS.UPDATE),
        for_resources=("collection",),
    )

    config.add_subscriber(
        functools.partial(
            listeners.create_editors_reviewers_groups,
            resources=resources,
            editors_group=global_settings["editors_group"],
            reviewers_group=global_settings["reviewers_group"],
        ),
        ResourceChanged,
        for_actions=(ACTIONS.CREATE,),
        for_resources=("collection",),
    )

    config.add_subscriber(
        functools.partial(listeners.cleanup_preview_destination, resources=resources),
        ResourceChanged,
        for_actions=(ACTIONS.DELETE,),
        for_resources=("collection",),
    )

    config.add_subscriber(
        functools.partial(listeners.prevent_collection_delete, resources=resources),
        ResourceChanged,
        for_actions=(ACTIONS.DELETE,),
        for_resources=("collection",),
    )

    if not asbool(settings.get("signer.allow_floats", False)):
        config.add_subscriber(
            functools.partial(listeners.prevent_float_value, resources=resources),
            ResourceChanged,
            for_actions=(ACTIONS.CREATE, ACTIONS.UPDATE),
            for_resources=("record",),
        )

    sign_data_listener = functools.partial(
        listeners.sign_collection_data, resources=resources, **global_settings
    )

    # If StatsD is enabled, monitor execution time of listener.
    if config.registry.statsd:
        # Due to https://github.com/jsocol/pystatsd/issues/85
        for attr in ("__module__", "__name__"):
            origin = getattr(listeners.sign_collection_data, attr)
            setattr(sign_data_listener, attr, origin)

        statsd_client = config.registry.statsd
        key = "plugins.signer"
        sign_data_listener = statsd_client.timer(key)(sign_data_listener)

    config.add_subscriber(
        sign_data_listener,
        ResourceChanged,
        for_actions=(ACTIONS.CREATE, ACTIONS.UPDATE),
        for_resources=("collection",),
    )

    def on_new_request(event):
        """Send the kinto-signer events in the before commit hook.
        This allows database operations done in subscribers to be automatically
        committed or rolledback.
        """
        # Since there is one transaction per batch, ignore subrequests.
        if hasattr(event.request, "parent"):
            return
        current = transaction.get()
        current.addBeforeCommitHook(listeners.send_signer_events, args=(event,))

    config.add_subscriber(on_new_request, NewRequest)

    try:
        from kinto_emailer import send_notification

        config.add_subscriber(send_notification, ReviewRequested)
        config.add_subscriber(send_notification, ReviewApproved)
        config.add_subscriber(send_notification, ReviewRejected)
    except ImportError:
        pass



import datetime
import unittest
from unittest import mock

import pytest
from kinto.core.storage.exceptions import RecordNotFoundError

from kinto_remote_settings.signer.updater import LocalUpdater
from kinto_remote_settings.signer.utils import STATUS

from .support import DummyRequest


class LocalUpdaterTest(unittest.TestCase):
    def setUp(self):
        self.storage = mock.MagicMock()
        self.permission = mock.MagicMock()
        self.signer_instance = mock.MagicMock()
        self.updater = LocalUpdater(
            source={"bucket": "sourcebucket", "collection": "sourcecollection"},
            destination={"bucket": "destbucket", "collection": "destcollection"},
            signer=self.signer_instance,
            storage=self.storage,
            permission=self.permission,
        )

        # Resource events are bypassed completely in this test suite.
        patcher = mock.patch("kinto_remote_settings.signer.utils.build_request")
        self.addCleanup(patcher.stop)
        patcher.start()

    def patch(self, obj, *args, **kwargs):
        patcher = mock.patch.object(obj, *args, **kwargs)
        self.addCleanup(patcher.stop)
        return patcher.start()

    def test_updater_raises_if_resources_are_not_set_properly(self):
        with pytest.raises(ValueError) as excinfo:
            LocalUpdater(
                source={"bucket": "source"},
                destination={},
                signer=self.signer_instance,
                storage=self.storage,
                permission=self.permission,
            )
        assert str(excinfo.value) == (
            "Resources should contain both " "bucket and collection"
        )

    def test_get_source_records_asks_storage_for_records(self):
        self.storage.list_all.return_value = []

        self.updater.get_source_records()
        self.storage.list_all.assert_called_with(
            resource_name="record",
            parent_id="/buckets/sourcebucket/collections/sourcecollection",
        )

    def test_get_destination_records(self):
        # We want to test get_destination_records with some records.
        records = [
            {"id": idx, "foo": "bar %s" % idx, "last_modified": 42 - idx}
            for idx in range(1, 4)
        ]
        self.storage.list_all.return_value = records
        self.updater.get_destination_records()
        self.storage.resource_timestamp.assert_called_with(
            resource_name="record",
            parent_id="/buckets/destbucket/collections/destcollection",
        )
        self.storage.list_all.assert_called_with(
            resource_name="record",
            parent_id="/buckets/destbucket/collections/destcollection",
        )

    def test_push_records_to_destination(self):
        self.patch(self.updater, "get_destination_records", return_value=([], 1324))
        records = [
            {"id": idx, "foo": "bar %s" % idx, "last_modified": 42 - idx}
            for idx in range(1, 4)
        ]
        self.patch(self.updater, "get_source_records", return_value=(records, 1325))
        self.updater.push_records_to_destination(DummyRequest())
        assert self.storage.update.call_count == 3
        assert self.storage.update.call_args_list[0][1] == {
            "obj": {"id": 1, "foo": "bar 1"},
            "object_id": 1,
            "parent_id": "/buckets/destbucket/collections/destcollection",
            "resource_name": "record",
        }

    def test_push_records_removes_deleted_records(self):
        self.patch(self.updater, "get_destination_records", return_value=([], 1324))
        records = [
            {"id": idx, "foo": "bar %s" % idx, "last_modified": 42 - idx}
            for idx in range(0, 2)
        ]
        records.extend(
            [{"id": idx, "deleted": True, "last_modified": 42} for idx in range(3, 5)]
        )
        self.patch(self.updater, "get_source_records", return_value=(records, 1325))
        self.updater.push_records_to_destination(DummyRequest())
        assert self.updater.get_source_records.call_count == 1
        assert self.storage.update.call_count == 2
        assert self.storage.delete.call_count == 2

    def test_push_records_skip_already_deleted_records(self):
        # In case the record doesn't exists in the destination
        # a RecordNotFoundError is raised.
        self.storage.delete.side_effect = RecordNotFoundError()
        self.patch(self.updater, "get_destination_records", return_value=([], 1324))
        records = [
            {"id": idx, "foo": "bar %s" % idx, "last_modified": 42 - idx}
            for idx in range(0, 2)
        ]
        records.extend(
            [{"id": idx, "deleted": True, "last_modified": 42} for idx in range(3, 5)]
        )
        self.patch(self.updater, "get_source_records", return_value=(records, 1325))
        # Calling the updater should not raise the RecordNotFoundError.
        self.updater.push_records_to_destination(DummyRequest())

    def test_push_records_to_destination_with_no_destination_changes(self):
        self.patch(self.updater, "get_destination_records", return_value=([], None))
        records = [
            {"id": idx, "foo": "bar %s" % idx, "last_modified": 42 - idx}
            for idx in range(1, 4)
        ]
        self.patch(self.updater, "get_source_records", return_value=(records, 1325))
        self.updater.push_records_to_destination(DummyRequest())
        assert self.updater.get_source_records.call_count == 1
        assert self.storage.update.call_count == 3

    def test_set_destination_signature_modifies_the_destination_collection(self):
        self.storage.get.return_value = {"id": 1234, "last_modified": 1234}
        self.updater.set_destination_signature(
            mock.sentinel.signature, {}, DummyRequest()
        )

        self.storage.update.assert_called_with(
            resource_name="collection",
            object_id="destcollection",
            parent_id="/buckets/destbucket",
            obj={"id": 1234, "signature": mock.sentinel.signature},
        )

    def test_set_destination_signature_copies_kinto_admin_ui_fields(self):
        self.storage.get.return_value = {
            "id": 1234,
            "sort": "-age",
            "last_modified": 1234,
        }
        self.updater.set_destination_signature(
            mock.sentinel.signature,
            {"displayFields": ["name"], "sort": "size"},
            DummyRequest(),
        )

        self.storage.update.assert_called_with(
            resource_name="collection",
            object_id="destcollection",
            parent_id="/buckets/destbucket",
            obj={
                "id": 1234,
                "signature": mock.sentinel.signature,
                "sort": "-age",
                "displayFields": ["name"],
            },
        )

    def test_update_source_status_modifies_the_source_collection(self):
        self.storage.get.return_value = {
            "id": 1234,
            "last_modified": 1234,
            "status": "to-sign",
        }

        with mock.patch("kinto_remote_settings.signer.updater.datetime") as mocked:
            mocked.datetime.now().isoformat.return_value = "2018-04-09"
            self.updater.update_source_status(STATUS.SIGNED, DummyRequest())

        self.storage.update.assert_called_with(
            resource_name="collection",
            object_id="sourcecollection",
            parent_id="/buckets/sourcebucket",
            obj={
                "id": 1234,
                "last_review_by": "basicauth:bob",
                "last_review_date": "2018-04-09",
                "last_signature_by": "basicauth:bob",
                "last_signature_date": "2018-04-09",
                "status": "signed",
            },
        )

    def test_create_destination_updates_collection_permissions(self):
        collection_uri = "/buckets/destbucket/collections/destcollection"
        request = DummyRequest()
        request.route_path.return_value = collection_uri
        self.updater.create_destination(request)
        request.registry.permission.replace_object_permissions.assert_called_with(
            collection_uri, {"read": ("system.Everyone",)}
        )

    def test_create_destination_creates_bucket(self):
        request = DummyRequest()
        self.updater.create_destination(request)
        request.registry.storage.create.assert_any_call(
            resource_name="bucket", parent_id="", obj={"id": "destbucket"}
        )

    def test_create_destination_creates_collection(self):
        bucket_id = "/buckets/destbucket"
        request = DummyRequest()
        self.updater.create_destination(request)
        request.registry.storage.create.assert_any_call(
            resource_name="collection",
            parent_id=bucket_id,
            obj={"id": "destcollection"},
        )

    def test_sign_and_update_destination(self):
        records = [
            {"id": idx, "foo": "bar %s" % idx, "last_modified": idx}
            for idx in range(1, 3)
        ]
        self.storage.list_all.return_value = records

        self.patch(self.storage, "update_records")
        self.patch(self.updater, "get_destination_records", return_value=([], "0"))
        self.patch(self.updater, "push_records_to_destination")
        self.patch(self.updater, "set_destination_signature")

        self.updater.sign_and_update_destination(DummyRequest(), {"id": "source"})

        assert self.updater.get_destination_records.call_count == 1
        assert self.updater.push_records_to_destination.call_count == 1
        assert self.updater.set_destination_signature.call_count == 1

    def test_refresh_signature_does_not_push_records(self):
        self.storage.list_all.return_value = []
        self.patch(self.updater, "set_destination_signature")
        self.patch(self.updater, "push_records_to_destination")

        self.updater.refresh_signature(DummyRequest(), "signed")

        assert self.updater.set_destination_signature.call_count == 1
        assert self.updater.push_records_to_destination.call_count == 0

    def test_refresh_signature_restores_status_on_source(self):
        self.storage.list_all.return_value = []
        with mock.patch("kinto_remote_settings.signer.updater.datetime") as mocked:
            mocked.datetime.now.return_value = datetime.datetime(2010, 10, 31)

            self.updater.refresh_signature(DummyRequest(), "work-in-progress")

        new_attrs = {
            "status": "work-in-progress",
            "last_signature_by": "basicauth:bob",
            "last_signature_date": "2010-10-31T00:00:00",
        }
        self.storage.update.assert_any_call(
            resource_name="collection",
            parent_id="/buckets/sourcebucket",
            object_id="sourcecollection",
            obj=new_attrs,
        )



import unittest

import pytest
from pyramid.exceptions import ConfigurationError

from kinto_remote_settings.signer import utils


class ParseResourcesTest(unittest.TestCase):
    def test_missing_arrow_raises_an_exception(self):
        raw_resources = """
        foo bar
        """
        with pytest.raises(ConfigurationError):
            utils.parse_resources(raw_resources)

    def test_non_local_first_argument_raises_an_exception(self):
        raw_resources = """
        foo -> bar
        bar -> baz
        """
        with pytest.raises(ConfigurationError):
            utils.parse_resources(raw_resources)

    def test_malformed_url_raises_an_exception(self):
        raw_resources = """
        /buckets/sbid/scid -> /buckets/dbid/collections/dcid
        """
        with pytest.raises(ConfigurationError):
            utils.parse_resources(raw_resources)

    def test_outnumbered_urls_raises_an_exception(self):
        raw_resources = (
            "/buckets/sbid/scid -> "
            "/buckets/dbid/collections/dcid -> "
            "/buckets/dbid/collections/dcid -> "
            "/buckets/sbid/scid"
        )
        with pytest.raises(ConfigurationError):
            utils.parse_resources(raw_resources)

    def test_returned_resources_match_the_expected_format(self):
        raw_resources = """
        /buckets/sbid/collections/scid -> /buckets/dbid/collections/dcid
        """
        resources = utils.parse_resources(raw_resources)
        assert resources == {
            "/buckets/sbid/collections/scid": {
                "source": {"bucket": "sbid", "collection": "scid"},
                "destination": {"bucket": "dbid", "collection": "dcid"},
            }
        }

    def test_returned_resources_match_the_legacy_format(self):
        raw_resources = """
        sbid/scid -> dbid/dcid
        """
        resources = utils.parse_resources(raw_resources)
        assert resources == {
            "/buckets/sbid/collections/scid": {
                "source": {"bucket": "sbid", "collection": "scid"},
                "destination": {"bucket": "dbid", "collection": "dcid"},
            }
        }

        raw_resources = """
        sbid/scid ; dbid/dcid
        """
        resources = utils.parse_resources(raw_resources)
        assert resources == {
            "/buckets/sbid/collections/scid": {
                "source": {"bucket": "sbid", "collection": "scid"},
                "destination": {"bucket": "dbid", "collection": "dcid"},
            }
        }

    def test_spaces_are_supported(self):
        raw_resources = """
        /buckets/bid1/collections/scid1 -> /buckets/bid1/collections/dcid1
        /buckets/bid2/collections/scid2 -> /buckets/bid2/collections/dcid2
        """
        resources = utils.parse_resources(raw_resources)
        assert len(resources) == 2
        assert (
            resources["/buckets/bid1/collections/scid1"]["source"]["bucket"] == "bid1"
        )
        assert (
            resources["/buckets/bid2/collections/scid2"]["source"]["bucket"] == "bid2"
        )

    def test_multiple_resources_are_supported(self):
        raw_resources = """
        /buckets/sbid1/collections/scid1 -> /buckets/dbid1/collections/dcid1
        /buckets/sbid2/collections/scid2 -> /buckets/dbid2/collections/dcid2
        """
        resources = utils.parse_resources(raw_resources)
        assert len(resources) == 2

    def test_a_preview_collection_is_supported(self):
        raw_resources = (
            "/buckets/stage/collections/cid -> "
            "/buckets/preview/collections/cid -> "
            "/buckets/prod/collections/cid -> "
        )
        resources = utils.parse_resources(raw_resources)
        assert resources == {
            "/buckets/stage/collections/cid": {
                "source": {"bucket": "stage", "collection": "cid"},
                "preview": {"bucket": "preview", "collection": "cid"},
                "destination": {"bucket": "prod", "collection": "cid"},
            }
        }

    def test_resources_should_be_space_separated(self):
        raw_resources = (
            "/buckets/sbid1/collections/scid -> /buckets/dbid1/collections/dcid,"
            "/buckets/sbid2/collections/scid -> /buckets/dbid2/collections/dcid"
        )
        with self.assertRaises(ConfigurationError):
            utils.parse_resources(raw_resources)

        raw_resources = "sbid1/scid -> dbid1/dcid,sbid2/scid -> dbid2/dcid"
        with self.assertRaises(ConfigurationError):
            utils.parse_resources(raw_resources)

    def test_resources_must_be_valid_names(self):
        raw_resources = (
            "/buckets/sbi+d1/collections/scid -> /buckets/dbid1/collections/dci,d"
        )
        with self.assertRaises(ConfigurationError) as e:
            utils.parse_resources(raw_resources)
        assert repr(e.exception).startswith(
            'ConfigurationError("Malformed resource: '
            "bucket or collection id is invalid"
        )

    def test_resources_can_be_defined_per_bucket(self):
        raw_resources = "/buckets/stage -> /buckets/preview -> /buckets/prod"
        resources = utils.parse_resources(raw_resources)
        assert resources == {
            "/buckets/stage": {
                "source": {"bucket": "stage", "collection": None},
                "preview": {"bucket": "preview", "collection": None},
                "destination": {"bucket": "prod", "collection": None},
            }
        }

    def test_cannot_mix_per_bucket_and_per_collection(self):
        raw_resources = "/buckets/stage -> /buckets/prod/collections/boom"
        with self.assertRaises(ConfigurationError):
            utils.parse_resources(raw_resources)

        raw_resources = (
            "/buckets/stage/collections/boom -> "
            "/buckets/preview/collections/boom -> "
            "/buckets/prod"
        )
        with self.assertRaises(ConfigurationError):
            utils.parse_resources(raw_resources)

        raw_resources = (
            "/buckets/stage -> /buckets/preview/collections/boom -> /buckets/prod"
        )
        with self.assertRaises(ConfigurationError):
            utils.parse_resources(raw_resources)

        raw_resources = "/buckets/stage/collections/boom -> /buckets/prod"
        with self.assertRaises(ConfigurationError):
            utils.parse_resources(raw_resources)

    def test_cannot_repeat_source_preview_or_destination(self):
        raw_resources = "/buckets/stage -> /buckets/stage -> /buckets/prod"
        with self.assertRaises(ConfigurationError):
            utils.parse_resources(raw_resources)

        raw_resources = "/buckets/stage -> /buckets/preview -> /buckets/stage"
        with self.assertRaises(ConfigurationError):
            utils.parse_resources(raw_resources)

        raw_resources = "/buckets/stage -> /buckets/preview -> /buckets/preview"
        with self.assertRaises(ConfigurationError):
            utils.parse_resources(raw_resources)

    def test_cannot_repeat_resources(self):
        # Repeated source.
        raw_resources = """
        /buckets/stage -> /buckets/preview1 -> /buckets/prod1
        /buckets/stage -> /buckets/preview2 -> /buckets/prod2
        """
        with self.assertRaises(ConfigurationError):
            utils.parse_resources(raw_resources)

        # Repeated reviews.
        raw_resources = """
        /buckets/stage1 -> /buckets/preview -> /buckets/prod1
        /buckets/stage2 -> /buckets/preview -> /buckets/prod2
        """
        with self.assertRaises(ConfigurationError):
            utils.parse_resources(raw_resources)

        # Repeated destination.
        raw_resources = """
        /buckets/stage1 -> /buckets/prod
        /buckets/stage2 -> /buckets/preview -> /buckets/prod
        """
        with self.assertRaises(ConfigurationError):
            utils.parse_resources(raw_resources)

        # Source in other's preview.
        raw_resources = """
        /buckets/stage -> /buckets/preview -> /buckets/prod
        /buckets/bid1  -> /buckets/stage   -> /buckets/bid2
        """
        with self.assertRaises(ConfigurationError):
            utils.parse_resources(raw_resources)

        # Source in other's destination.
        raw_resources = """
    /buckets/b/collections/c  -> /buckets/b/collections/c2 -> /buckets/b/collections/c3
    /buckets/b/collections/ca -> /buckets/b/collections/cb -> /buckets/b/collections/c
        """
        with self.assertRaises(ConfigurationError):
            utils.parse_resources(raw_resources)

        # Preview in other's destination.
        raw_resources = """
    /buckets/b/collections/c0 -> /buckets/b/collections/c1 -> /buckets/b/collections/c2
    /buckets/b/collections/ca -> /buckets/b/collections/cb -> /buckets/b/collections/c1
        """
        with self.assertRaises(ConfigurationError):
            utils.parse_resources(raw_resources)



import base64
import warnings
from urllib.parse import urljoin

import requests
from kinto import logger
from requests_hawk import HawkAuth

from ..utils import get_first_matching_setting
from .base import SignerBase

SIGNATURE_FIELDS = ["signature", "x5u"]
EXTRA_SIGNATURE_FIELDS = ["mode", "public_key", "type", "signer_id", "ref"]


class AutographSigner(SignerBase):
    def __init__(self, server_url, hawk_id, hawk_secret):
        self.server_url = server_url
        self.auth = HawkAuth(id=hawk_id, key=hawk_secret)

    def sign(self, payload):
        if isinstance(payload, str):  # pragma: nocover
            payload = payload.encode("utf-8")

        b64_payload = base64.b64encode(payload)
        url = urljoin(self.server_url, "/sign/data")
        resp = requests.post(
            url, auth=self.auth, json=[{"input": b64_payload.decode("utf-8")}]
        )
        resp.raise_for_status()
        signature_bundle = resp.json()[0]

        # Critical fields must be present, will raise if missing.
        infos = {field: signature_bundle[field] for field in SIGNATURE_FIELDS}
        # Other fields are returned and will be stored as part of the signature.
        # but client won't break if they are missing, so don't raise.
        infos.update(
            **{
                field: signature_bundle[field]
                for field in EXTRA_SIGNATURE_FIELDS
                if field in signature_bundle
            }
        )
        logger.info(
            "Obtained %s response from Autograph %s"
            % (resp.status_code, signature_bundle["ref"])
        )
        return infos


def load_from_settings(settings, prefix="", *, prefixes=None):
    if prefixes is None:
        prefixes = [prefix]

    if prefix != "":
        message = (
            "signer.load_from_settings `prefix` parameter is deprecated, please "
            "use `prefixes` instead."
        )
        warnings.warn(message, DeprecationWarning)

    return AutographSigner(
        server_url=get_first_matching_setting(
            "autograph.server_url", settings, prefixes
        ),
        hawk_id=get_first_matching_setting("autograph.hawk_id", settings, prefixes),
        hawk_secret=get_first_matching_setting(
            "autograph.hawk_secret", settings, prefixes
        ),
    )



import tempfile
import unittest

from kinto_remote_settings.signer.backends.local_ecdsa import ECDSASigner
from kinto_remote_settings.signer.generate_keypair import generate_keypair


class KeyPairGeneratorTest(unittest.TestCase):
    def test_generated_keypairs_can_be_loaded(self):
        private_key_location = tempfile.mktemp("private_key")
        public_key_location = tempfile.mktemp("public_key")

        generate_keypair(private_key_location, public_key_location)
        backend = ECDSASigner(private_key=private_key_location)
        backend.sign("test")



from collections import OrderedDict
from enum import Enum

from kinto.core.events import ACTIONS
from kinto.core.storage.exceptions import UnicityError
from kinto.core.utils import build_request, instance_uri
from kinto.views import NameGenerator
from pyramid.exceptions import ConfigurationError

PLUGIN_USERID = "plugin:kinto-signer"
FIELD_LAST_MODIFIED = "last_modified"


class STATUS(Enum):
    WORK_IN_PROGRESS = "work-in-progress"
    TO_SIGN = "to-sign"
    TO_REFRESH = "to-resign"
    TO_REVIEW = "to-review"
    TO_ROLLBACK = "to-rollback"
    SIGNED = "signed"

    def __eq__(self, other):
        if not hasattr(other, "value"):
            return self.value == other
        return super(STATUS, self).__eq__(other)

    def __ne__(self, other):
        return not self.__eq__(other)


def _get_resource(resource):
    # Use the default NameGenerator in Kinto resources to check if the resource
    # URIs seem valid.
    # XXX: if a custom ID generator is specified in settings, this verification would
    # not result as expected.
    name_generator = NameGenerator()

    parts = resource.split("/")
    if len(parts) == 2:
        bucket, collection = parts
    elif len(parts) == 3 and parts[1] == "buckets":
        # /buckets/bid
        _, _, bucket = parts
        collection = None
    elif len(parts) == 5 and parts[1] == "buckets" and parts[3] == "collections":
        # /buckets/bid/collections/cid
        _, _, bucket, _, collection = parts
    else:
        raise ValueError("should be a bucket or collection URI")
    valid_ids = name_generator.match(bucket) and (
        collection is None or name_generator.match(collection)
    )
    if not valid_ids:
        raise ValueError("bucket or collection id is invalid")
    return {"bucket": bucket, "collection": collection}


def parse_resources(raw_resources):
    resources = OrderedDict()

    lines = [line.strip() for line in raw_resources.strip().splitlines()]
    for res in lines:
        error_msg = "Malformed resource: %%s (in %r). See kinto-signer README." % res
        if "->" not in res and ";" not in res:
            raise ConfigurationError(error_msg % "not separated with '->'")

        try:
            triplet = [
                r.strip() for r in res.replace(";", " ").replace("->", " ").split()
            ]
            if len(triplet) == 2:
                source_uri, destination_uri = triplet
                preview_uri = None
            else:
                source_uri, preview_uri, destination_uri = triplet
        except ValueError:
            raise ConfigurationError(error_msg % "should be a pair or a triplet")

        try:
            source = _get_resource(source_uri)
            destination = _get_resource(destination_uri)
            preview = _get_resource(preview_uri) if preview_uri else None
        except ValueError as e:
            raise ConfigurationError(error_msg % e)

        # Raise if mix-up of per-bucket/specific collection.
        sections = (source, destination) + ((preview,) if preview else tuple())
        all_per_bucket = all([x["collection"] is None for x in sections])
        all_explicit = all([x["collection"] is not None for x in sections])
        if not all_per_bucket and not all_explicit:
            raise ConfigurationError(
                error_msg % "cannot mix bucket and collection URIs"
            )

        # Repeated source/preview/destination.
        if (
            len(set([tuple(s.items()) for s in (source, preview or {}, destination)]))
            != 3
        ):
            raise ConfigurationError(
                error_msg % "cannot have same value for source, "
                " preview or destination"
            )

        # Resources info is returned as a mapping by bucket/collection URI.
        bid = source["bucket"]
        if source["collection"] is None:
            # Per bucket.
            key = f"/buckets/{bid}"
        else:
            cid = source["collection"]
            # For a specific collection.
            key = f"/buckets/{bid}/collections/{cid}"

        # We can't have the same source twice.
        if key in resources:
            raise ConfigurationError(error_msg % "cannot repeat resource")

        resources[key] = {"source": source, "destination": destination}
        if preview is not None:
            resources[key]["preview"] = preview

    # Raise if same bid/cid twice/thrice.
    # Theoretically we could support it, but since we never said it was possible
    # and have no test at all for that, prefer safety.
    sources = [tuple(r["source"].items()) for r in resources.values()]
    destinations = [tuple(r["destination"].items()) for r in resources.values()]
    previews = [
        tuple(r["preview"].items()) for r in resources.values() if "preview" in r
    ]

    if len(set(destinations)) != len(destinations):
        raise ConfigurationError("Resources setting has repeated destination URI")
    if len(set(previews)) != len(previews):
        raise ConfigurationError("Resources setting has repeated preview URI")

    intersects = (
        set(sources).intersection(set(previews))
        or set(sources).intersection(set(destinations))
        or set(destinations).intersection(set(previews))
    )
    if intersects:
        raise ConfigurationError("cannot repeat URIs across resources")

    return resources


def get_first_matching_setting(setting_name, settings, prefixes, default=None):
    for prefix in prefixes:
        prefixed_setting_name = prefix + setting_name
        if prefixed_setting_name in settings:
            return settings[prefixed_setting_name]
    return default


def ensure_resource_exists(
    request, resource_name, parent_id, obj, permissions, matchdict
):
    storage = request.registry.storage
    permission = request.registry.permission
    try:
        created = storage.create(
            resource_name=resource_name, parent_id=parent_id, obj=obj
        )
        object_uri = instance_uri(request, resource_name, **matchdict)
        permission.replace_object_permissions(object_uri, permissions)
        notify_resource_event(
            request,
            {"method": "PUT", "path": object_uri},
            matchdict=matchdict,
            resource_name=resource_name,
            parent_id=parent_id,
            obj=created,
            action=ACTIONS.CREATE,
        )
    except UnicityError:
        pass


def notify_resource_event(
    request, request_options, matchdict, resource_name, parent_id, obj, action, old=None
):
    """Helper that triggers resource events as real requests."""
    fakerequest = build_request(request, request_options)
    fakerequest.matchdict = matchdict
    fakerequest.bound_data = request.bound_data
    fakerequest.authn_type, fakerequest.selected_userid = PLUGIN_USERID.split(":")
    fakerequest.current_resource_name = resource_name

    # When kinto-signer copies record from one place to another,
    # it simulates a resource event. Since kinto-attachment
    # prevents from updating attachment fields, it throws an error.
    # The following flag will disable the kinto-attachment check.
    # See https://github.com/Kinto/kinto-signer/issues/256
    # and https://bugzilla.mozilla.org/show_bug.cgi?id=1470812
    has_changed_attachment = (
        resource_name == "record"
        and action == ACTIONS.UPDATE
        and "attachment" in old
        and old["attachment"] != obj.get("attachment")
    )
    if has_changed_attachment:
        fakerequest._attachment_auto_save = True

    fakerequest.notify_resource_event(
        parent_id=parent_id,
        timestamp=obj[FIELD_LAST_MODIFIED],
        data=obj,
        action=action,
        old=old,
    )


def records_equal(a, b):
    ignore_fields = ("last_modified", "schema")
    ac = {k: v for k, v in a.items() if k not in ignore_fields}
    bc = {k: v for k, v in b.items() if k not in ignore_fields}
    return ac == bc


def records_diff(left, right):
    left_by_id = {r["id"]: r for r in left}
    results = []
    for r in right:
        rid = r["id"]
        left_record = left_by_id.pop(rid, None)
        if left_record is None:
            # In right, but not in left (deleted!)
            results.append({**r, "deleted": True})
        elif not records_equal(left_record, r):
            # Differ between left and right
            results.append(left_record)
    # In left, but not in right.
    results.extend(left_by_id.values())
    return results



from pyramid.settings import aslist

from .. import __version__

MONITOR_BUCKET = "monitor"
MONITOR_BUCKET_PATH = "/buckets/{}".format(MONITOR_BUCKET)
CHANGES_COLLECTION = "changes"
CHANGES_COLLECTION_PATH = "{}/collections/{}".format(
    MONITOR_BUCKET_PATH, CHANGES_COLLECTION
)
CHANGES_RECORDS_PATH = "{}/records".format(CHANGES_COLLECTION_PATH)
CHANGESET_PATH = "/buckets/{bid}/collections/{cid}/changeset"


def includeme(config):
    settings = config.get_settings()
    collections = settings.get("changes.resources", [])

    config.add_api_capability(
        "changes",
        description="Track modifications of records in Kinto and store"
        " the collection timestamps into a specific bucket"
        " and collection.",
        url="http://kinto.readthedocs.io/en/latest/tutorials/"
        "synchronisation.html#polling-for-remote-changes",
        version=__version__,
        collections=aslist(collections),
    )

    config.scan("kinto_remote_settings.changes.views")



import hashlib
from uuid import UUID

from kinto.core import utils as core_utils
from pyramid.settings import aslist


def monitored_collections(registry):
    storage = registry.storage
    resources_uri = aslist(registry.settings.get("changes.resources", ""))
    collections = []

    for resource_uri in resources_uri:
        resource_name, matchdict = core_utils.view_lookup_registry(
            registry, resource_uri
        )
        if resource_name == "bucket":
            # Every collections in this bucket.
            result = storage.list_all(
                resource_name="collection", parent_id=resource_uri
            )
            collections.extend([(matchdict["id"], obj["id"]) for obj in result])

        elif resource_name == "collection":
            collections.append((matchdict["bucket_id"], matchdict["id"]))

    return collections


def changes_object(request, bucket_id, collection_id, timestamp):
    """Generate an object for /buckets/monitor/collections/changes."""
    http_host = request.registry.settings.get("http_host") or ""
    collection_uri = core_utils.instance_uri(
        request, "collection", bucket_id=bucket_id, id=collection_id
    )
    uniqueid = http_host + collection_uri
    identifier = hashlib.md5(uniqueid.encode("utf-8")).hexdigest()
    entry_id = str(UUID(identifier))

    return dict(
        id=entry_id,
        last_modified=timestamp,
        bucket=bucket_id,
        collection=collection_id,
        host=http_host,
    )



import os
import tempfile
import unittest
from base64 import urlsafe_b64decode, urlsafe_b64encode
from unittest import mock

import pytest

from kinto_remote_settings.signer.backends import (
    autograph,
    base,
    exceptions,
    local_ecdsa,
)

SIGNATURE = (
    "ikfq6qOV85vR7QaNCTldVvvtcNpPIICqqMp3tfyiT7fHCgFNq410SFnIfjAPgSa"
    "jEtxxyGtZFMoI_BzO_1y5oShLtX0LH4wx_Wft7wz17T7fFqpDQ9hFZzTOPBwZUIbx"
)


def save_key(key, key_name):
    tmp = tempfile.mktemp(key_name)
    with open(tmp, "wb+") as tmp_file:
        tmp_file.write(key)
    return tmp


class BaseSignerTest(unittest.TestCase):
    def test_base_method_raises_unimplemented(self):
        signer = base.SignerBase()
        with pytest.raises(NotImplementedError):
            signer.sign("TEST")


class ECDSASignerTest(unittest.TestCase):
    @classmethod
    def get_backend(cls, **options):
        return local_ecdsa.ECDSASigner(**options)

    @classmethod
    def setUpClass(cls):
        sk, vk = local_ecdsa.ECDSASigner.generate_keypair()
        cls.sk_location = save_key(sk, "signing-key")
        cls.vk_location = save_key(vk, "verifying-key")
        cls.signer = cls.get_backend(private_key=cls.sk_location)

    @classmethod
    def tearDownClass(cls):
        os.remove(cls.sk_location)
        os.remove(cls.vk_location)

    def test_keyloading_fails_if_no_settings(self):
        backend = self.get_backend(public_key=self.vk_location)
        with pytest.raises(ValueError):
            backend.load_private_key()

    def test_key_loading_works(self):
        key = self.signer.load_private_key()
        assert key is not None

    def test_signer_roundtrip(self):
        signature = self.signer.sign("this is some text")
        self.signer.verify("this is some text", signature)

    def test_base64url_encoding(self):
        signature_bundle = self.signer.sign("this is some text")
        b64signature = signature_bundle["signature"]

        decoded_signature = urlsafe_b64decode(b64signature.encode("utf-8"))
        b64urlsignature = urlsafe_b64encode(decoded_signature).decode("utf-8")
        signature_bundle["signature"] = b64urlsignature
        signature_bundle["signature_encoding"] = "rs_base64url"

        self.signer.verify("this is some text", signature_bundle)

    def test_wrong_signature_raises_an_error(self):
        signature_bundle = {"signature": SIGNATURE, "mode": "p384ecdsa", "ref": ""}

        with pytest.raises(exceptions.BadSignatureError):
            self.signer.verify("Text not matching with the sig.", signature_bundle)

    def test_signer_returns_a_base64_string(self):
        signature = self.signer.sign("this is some text")["signature"]
        urlsafe_b64decode(signature.encode("utf-8"))  # Raise if wrong.

    def test_load_private_key_raises_if_no_key_specified(self):
        with pytest.raises(ValueError):
            self.get_backend().load_private_key()

    def test_public_key_can_be_loaded_from_public_key_pem(self):
        signer = self.get_backend(public_key=self.vk_location)
        signer.load_public_key()

    def test_public_key_can_be_loaded_from_private_key_pem(self):
        signer = self.get_backend(private_key=self.sk_location)
        signer.load_public_key()

    def test_load_public_key_raises_an_error_if_missing_settings(self):
        with pytest.raises(ValueError) as excinfo:
            self.get_backend()
        msg = "Please, specify either a private_key or public_key location."
        assert str(excinfo.value) == msg

    @mock.patch("kinto_remote_settings.signer.backends.local_ecdsa.ECDSASigner")
    def test_load_from_settings(self, mocked_signer):
        local_ecdsa.load_from_settings(
            {
                "signer.ecdsa.private_key": mock.sentinel.private_key,
                "signer.ecdsa.public_key": mock.sentinel.public_key,
            },
            prefix="signer.",
        )

        mocked_signer.assert_called_with(
            private_key=mock.sentinel.private_key, public_key=mock.sentinel.public_key
        )

    def test_load_from_settings_fails_if_no_public_or_private_key(self):
        with pytest.raises(ValueError) as excinfo:
            local_ecdsa.load_from_settings({}, "")
        msg = (
            "Please specify either kinto.signer.ecdsa.private_key or "
            "kinto.signer.ecdsa.public_key in the settings."
        )
        assert str(excinfo.value) == msg


class AutographSignerTest(unittest.TestCase):
    def setUp(self):
        self.signer = autograph.AutographSigner(
            hawk_id="alice",
            hawk_secret="fs5wgcer9qj819kfptdlp8gm227ewxnzvsuj9ztycsx08hfhzu",
            server_url="http://localhost:8000",
        )

    @mock.patch("kinto_remote_settings.signer.backends.autograph.requests")
    def test_request_is_being_crafted_with_payload_as_input(self, requests):
        response = mock.MagicMock()
        response.json.return_value = [{"signature": SIGNATURE, "x5u": "", "ref": ""}]
        requests.post.return_value = response
        signature_bundle = self.signer.sign("test data")
        requests.post.assert_called_with(
            "http://localhost:8000/sign/data",
            auth=self.signer.auth,
            json=[{"input": "dGVzdCBkYXRh"}],
        )
        assert signature_bundle["signature"] == SIGNATURE

    @mock.patch("kinto_remote_settings.signer.backends.autograph.AutographSigner")
    def test_load_from_settings(self, mocked_signer):
        autograph.load_from_settings(
            {
                "signer.autograph.server_url": mock.sentinel.server_url,
                "signer.autograph.hawk_id": mock.sentinel.hawk_id,
                "signer.autograph.hawk_secret": mock.sentinel.hawk_secret,
            },
            prefix="signer.",
        )

        mocked_signer.assert_called_with(
            server_url=mock.sentinel.server_url,
            hawk_id=mock.sentinel.hawk_id,
            hawk_secret=mock.sentinel.hawk_secret,
        )



import unittest
import uuid
from unittest import mock

import pytest
from kinto import main as kinto_main
from kinto.core.events import ResourceChanged
from pyramid import testing
from pyramid.exceptions import ConfigurationError
from requests import exceptions as requests_exceptions

from kinto_remote_settings import __version__
from kinto_remote_settings.signer import includeme, utils
from kinto_remote_settings.signer.backends.autograph import AutographSigner
from kinto_remote_settings.signer.listeners import sign_collection_data

from .support import BaseWebTest, get_user_headers


class HelloViewTest(BaseWebTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super().get_app_settings(extras)
        settings["signer.reviewers_group"] = "{bucket_id}-{collection_id}-reviewers"
        settings["signer.alice.reviewers_group"] = "revoyeurs"
        settings["signer.alice.source.to_review_enabled"] = "true"

        settings["signer.stage.normandy.to_review_enabled"] = "false"
        return settings

    def test_capability_is_exposed(self):
        self.maxDiff = None
        resp = self.app.get("/")
        capabilities = resp.json["capabilities"]
        self.assertIn("signer", capabilities)
        expected = {
            "description": "Digital signatures for integrity and authenticity of records.",  # NOQA
            "url": ("https://github.com/Kinto/kinto-signer#kinto-signer"),
            "version": __version__,
            "group_check_enabled": True,
            "to_review_enabled": False,
            "editors_group": "editors",
            "reviewers_group": "{bucket_id}-{collection_id}-reviewers",
            "resources": [
                {
                    "destination": {"bucket": "alice", "collection": "destination"},
                    "source": {"bucket": "alice", "collection": "source"},
                    "reviewers_group": "revoyeurs",
                    "to_review_enabled": True,
                },
                {
                    "destination": {"bucket": "alice", "collection": "to"},
                    "preview": {"bucket": "alice", "collection": "preview"},
                    "source": {"bucket": "alice", "collection": "from"},
                    "reviewers_group": "revoyeurs",
                },
                {
                    "destination": {"bucket": "bob", "collection": "destination"},
                    "source": {"bucket": "bob", "collection": "source"},
                    "reviewers_group": "bob-source-reviewers",
                },
                {
                    "source": {"bucket": "stage", "collection": None},
                    "preview": {"bucket": "preview", "collection": None},
                    "destination": {"bucket": "prod", "collection": None},
                    "reviewers_group": "stage-{collection_id}-reviewers",
                },
                {
                    "source": {"bucket": "stage", "collection": "normandy"},
                    "preview": {"bucket": "preview", "collection": "normandy"},
                    "destination": {"bucket": "prod", "collection": "normandy"},
                    "reviewers_group": "stage-normandy-reviewers",
                },
            ],
        }
        self.assertEqual(expected, capabilities["signer"])


class HeartbeatTest(BaseWebTest, unittest.TestCase):
    def setUp(self):
        patch = mock.patch("kinto_remote_settings.signer.backends.autograph.requests")
        self.mock = patch.start()
        self.addCleanup(patch.stop)
        self.signature = {"signature": "", "x5u": "", "mode": "", "ref": "abc"}
        self.mock.post.return_value.json.return_value = [self.signature]

    def test_heartbeat_is_exposed(self):
        resp = self.app.get("/__heartbeat__")
        assert "signer" in resp.json

    def test_heartbeat_fails_if_unreachable(self):
        self.mock.post.side_effect = requests_exceptions.ConnectTimeout()
        resp = self.app.get("/__heartbeat__", status=503)
        assert resp.json["signer"] is False

    def test_heartbeat_fails_if_missing_attributes(self):
        invalid = self.signature.copy()
        invalid.pop("signature")
        self.mock.post.return_value.json.return_value = [invalid]
        resp = self.app.get("/__heartbeat__", status=503)
        assert resp.json["signer"] is False


class IncludeMeTest(unittest.TestCase):
    def includeme(self, settings):
        config = testing.setUp(settings=settings)
        kinto_main(None, config=config)
        includeme(config)
        return config

    def test_includeme_raises_value_error_if_no_resource_defined(self):
        with pytest.raises(ConfigurationError):
            self.includeme(
                settings={"signer.ecdsa.private_key": "", "signer.ecdsa.public_key": ""}
            )

    def test_defines_a_signer_per_bucket(self):
        settings = {
            "signer.resources": (
                "/buckets/sb1/collections/sc1 -> /buckets/db1/collections/dc1\n"
            ),
            "signer.sb1.signer_backend": "kinto_remote_settings.signer.backends.local_ecdsa",  # noqa: 501
            "signer.sb1.ecdsa.public_key": "/path/to/key",
            "signer.sb1.ecdsa.private_key": "/path/to/private",
        }
        config = self.includeme(settings)
        (signer,) = config.registry.signers.values()
        assert signer.public_key == "/path/to/key"

    def test_includeme_doesnt_fail_when_expanding_collection(self):
        settings = {
            "signer.resources": (
                "/buckets/sb1 -> /buckets/db1\n" "/buckets/sb2 -> /buckets/db2\n"
            ),
            "signer.signer_backend": "kinto_remote_settings.signer.backends.local_ecdsa",
            "signer.ecdsa.public_key": "/path/to/key",
            "signer.ecdsa.private_key": "/path/to/private",
            "signer.sb1.sc1.signer_backend": "kinto_remote_settings.signer.backends.local_ecdsa",  # noqa: 501
            "signer.sb1.sc1.ecdsa.public_key": "/path/to/key",
            "signer.sb1.sc1.ecdsa.private_key": "/path/to/private",
        }
        self.includeme(settings)

    def test_includeme_sanitizes_exposed_settings(self):
        settings = {
            "signer.resources": (
                "/buckets/sb1 -> /buckets/db1\n" "/buckets/sb2 -> /buckets/db2\n"
            ),
            "signer.signer_backend": "kinto_remote_settings.signer.backends.local_ecdsa",
            "signer.ecdsa.public_key": "/path/to/key",
            "signer.ecdsa.private_key": "/path/to/private",
            "signer.sb1.sc1.signer_backend": "kinto_remote_settings.signer.backends.local_ecdsa",  # noqa: 501
            "signer.sb1.sc1.ecdsa.public_key": "/path/to/key",
            "signer.sb1.sc1.ecdsa.private_key": "/path/to/private",
            "signer.sb2.signer_backend": "kinto_remote_settings.signer.backends.local_ecdsa",  # noqa: 501
            "signer.sb2.ecdsa.public_key": "/path/to/key",
            "signer.sb2.ecdsa.private_key": "/path/to/private",
        }
        config = self.includeme(settings)
        all_capabilities = config.registry.api_capabilities
        capabilities = all_capabilities["signer"]
        for resource in capabilities["resources"]:
            assert "ecdsa.private_key" not in resource
            assert "private_key" not in resource

    def test_defines_a_signer_per_bucket_and_collection(self):
        settings = {
            "signer.resources": (
                "/buckets/sb1/collections/sc1 -> /buckets/db1/collections/dc1\n"
                "/buckets/sb1/collections/sc2 -> /buckets/db1/collections/dc2"
            ),
            "signer.sb1.signer_backend": "kinto_remote_settings.signer.backends.local_ecdsa",  # noqa: 501
            "signer.sb1.ecdsa.public_key": "/path/to/key",
            "signer.sb1.ecdsa.private_key": "/path/to/private",
            "signer.sb1.sc1.signer_backend": "kinto_remote_settings.signer.backends.autograph",  # noqa: 501
            "signer.sb1.sc1.autograph.server_url": "http://localhost",
            "signer.sb1.sc1.autograph.hawk_id": "alice",
            "signer.sb1.sc1.autograph.hawk_secret": "a-secret",
        }
        config = self.includeme(settings)
        signer1, signer2 = config.registry.signers.values()
        if isinstance(signer1, AutographSigner):
            signer1, signer2 = signer2, signer1
        assert signer1.public_key == "/path/to/key"
        assert signer2.server_url == "http://localhost"

    def test_falls_back_to_global_settings_if_not_defined(self):
        settings = {
            "signer.resources": (
                "/buckets/sb1/collections/sc1 -> /buckets/db1/collections/dc1\n"
                "/buckets/sb1/collections/sc2 -> /buckets/db1/collections/dc2"
            ),
            "signer.signer_backend": "kinto_remote_settings.signer.backends.autograph",
            "signer.autograph.server_url": "http://localhost",
            "signer.sb1.autograph.hawk_id": "bob",
            "signer.sb1.autograph.hawk_secret": "a-secret",
            "signer.sb1.sc1.autograph.hawk_id": "alice",
            "signer.sb1.sc1.autograph.hawk_secret": "a-secret",
        }
        config = self.includeme(settings)

        signer1 = config.registry.signers["/buckets/sb1/collections/sc1"]
        assert isinstance(signer1, AutographSigner)
        assert signer1.server_url == "http://localhost"
        assert signer1.auth.credentials["id"] == "alice"

        signer2 = config.registry.signers["/buckets/sb1/collections/sc2"]
        assert isinstance(signer2, AutographSigner)
        assert signer2.server_url == "http://localhost"
        assert signer2.auth.credentials["id"] == "bob"

    def test_a_statsd_timer_is_used_for_signature_if_configured(self):
        settings = {
            "statsd_url": "udp://127.0.0.1:8125",
            "signer.resources": (
                "/buckets/sb1/collections/sc1 -> /buckets/db1/collections/dc1"
            ),
            "signer.ecdsa.public_key": "/path/to/key",
            "signer.ecdsa.private_key": "/path/to/private",
        }
        config = self.includeme(settings)

        payload = dict(resource_name="collection", action="update", bucket_id="foo")
        event = ResourceChanged(
            payload=payload, impacted_objects=[], request=mock.MagicMock()
        )
        statsd_client = config.registry.statsd._client
        with mock.patch.object(statsd_client, "timing") as mocked:
            config.registry.notify(event)
            timers = set(c[0][0] for c in mocked.call_args_list)
            assert "plugins.signer" in timers

    def test_includeme_raises_value_error_if_unknown_placeholder(self):
        settings = {
            "signer.resources": "/buckets/sb1/collections/sc1 -> /buckets/db1/collections/dc1",  # noqa: 501
            "signer.editors_group": "{datetime}_group",
            "signer.ecdsa.public_key": "/path/to/key",
            "signer.ecdsa.private_key": "/path/to/private",
        }
        with pytest.raises(ConfigurationError) as excinfo:
            self.includeme(settings=settings)
        assert "Unknown group placeholder 'datetime'" in repr(excinfo.value)


class OnCollectionChangedTest(unittest.TestCase):
    def setUp(self):
        patch = mock.patch("kinto_remote_settings.signer.listeners.LocalUpdater")
        self.updater_mocked = patch.start()
        self.addCleanup(patch.stop)

    def test_nothing_happens_when_resource_is_not_configured(self):
        evt = mock.MagicMock(
            payload={"action": "update", "bucket_id": "a", "collection_id": "b"}
        )
        sign_collection_data(
            evt, resources=utils.parse_resources("c/d -> e/f"), to_review_enabled=True
        )
        assert not self.updater_mocked.called

    def test_nothing_happens_when_status_is_not_to_sign(self):
        evt = mock.MagicMock(
            payload={"action": "update", "bucket_id": "a", "collection_id": "b"},
            impacted_objects=[{"new": {"id": "b", "status": "signed"}}],
        )
        sign_collection_data(
            evt, resources=utils.parse_resources("a/b -> c/d"), to_review_enabled=True
        )
        assert not self.updater_mocked.sign_and_update_destination.called

    def test_updater_is_called_when_resource_and_status_matches(self):
        evt = mock.MagicMock(
            payload={"action": "update", "bucket_id": "a", "collection_id": "b"},
            impacted_objects=[{"new": {"id": "b", "status": "to-sign"}}],
        )
        evt.request.registry.storage = mock.sentinel.storage
        evt.request.registry.permission = mock.sentinel.permission
        evt.request.registry.signers = {
            "/buckets/a/collections/b": mock.sentinel.signer
        }
        evt.request.route_path.return_value = "/v1/buckets/a/collections/b"
        sign_collection_data(
            evt, resources=utils.parse_resources("a/b -> c/d"), to_review_enabled=True
        )
        self.updater_mocked.assert_called_with(
            signer=mock.sentinel.signer,
            storage=mock.sentinel.storage,
            permission=mock.sentinel.permission,
            source={"bucket": "a", "collection": "b"},
            destination={"bucket": "c", "collection": "d"},
        )

        mocked = self.updater_mocked.return_value
        assert mocked.sign_and_update_destination.called

    def test_kinto_attachment_property_is_set_to_allow_metadata_updates(self):
        evt = mock.MagicMock(
            payload={"action": "update", "bucket_id": "a", "collection_id": "b"},
            impacted_objects=[{"new": {"id": "b", "status": "to-sign"}}],
        )
        evt.request.registry.storage = mock.sentinel.storage
        evt.request.registry.permission = mock.sentinel.permission
        evt.request.registry.signers = {
            "/buckets/a/collections/b": mock.sentinel.signer
        }
        evt.request.route_path.return_value = "/v1/buckets/a/collections/b"
        sign_collection_data(
            evt, resources=utils.parse_resources("a/b -> c/d"), to_review_enabled=True
        )
        assert evt.request._attachment_auto_save is True

    def test_updater_does_not_fail_when_payload_is_inconsistent(self):
        # This happens with events on default bucket for kinto < 3.3
        evt = mock.MagicMock(
            payload={"action": "update", "subpath": "collections/boom"}
        )
        sign_collection_data(
            evt, resources=utils.parse_resources("a/b -> c/d"), to_review_enabled=True
        )


class BatchTest(BaseWebTest, unittest.TestCase):
    def setUp(self):
        super(BatchTest, self).setUp()
        self.headers = get_user_headers("me")

        resp = self.app.get("/", headers=self.headers)
        self.userid = resp.json["user"]["id"]

        self.app.put_json("/buckets/alice", headers=self.headers)
        self.app.put_json("/buckets/bob", headers=self.headers)

        self.app.put_json(
            "/buckets/alice/groups/reviewers",
            {"data": {"members": [self.userid]}},
            headers=self.headers,
        )
        self.app.put_json(
            "/buckets/bob/groups/reviewers",
            {"data": {"members": [self.userid]}},
            headers=self.headers,
        )

        # Patch calls to Autograph.
        patch = mock.patch("kinto_remote_settings.signer.backends.autograph.requests")
        self.mock = patch.start()
        self.addCleanup(patch.stop)
        self.mock.post.return_value.json.return_value = [
            {
                "signature": "",
                "hash_algorithm": "",
                "signature_encoding": "",
                "content-signature": "",
                "x5u": "",
                "ref": "",
            }
        ]

    def test_various_collections_can_be_signed_using_batch(self):
        self.app.put_json("/buckets/alice/collections/source", headers=self.headers)
        self.app.post_json(
            "/buckets/alice/collections/source/records", headers=self.headers
        )
        self.app.put_json("/buckets/bob/collections/source", headers=self.headers)
        self.app.post_json(
            "/buckets/bob/collections/source/records", headers=self.headers
        )

        self.app.post_json(
            "/batch",
            {
                "defaults": {
                    "method": "PATCH",
                    "body": {"data": {"status": "to-sign"}},
                },
                "requests": [
                    {"path": "/buckets/alice/collections/source"},
                    {"path": "/buckets/bob/collections/source"},
                ],
            },
            headers=self.headers,
        )

        resp = self.app.get("/buckets/alice/collections/source", headers=self.headers)
        assert resp.json["data"]["status"] == "signed"
        resp = self.app.get("/buckets/bob/collections/source", headers=self.headers)
        assert resp.json["data"]["status"] == "signed"

    def test_various_collections_can_be_signed_using_batch_creation(self):
        self.app.post_json(
            "/batch",
            {
                "defaults": {"method": "POST", "path": "/buckets/alice/collections"},
                "requests": [
                    {"body": {"data": {"id": "source", "status": "to-sign"}}},
                    {"body": {"data": {"id": "ignored", "status": "to-sign"}}},
                    {"body": {"data": {"id": "from", "status": "to-sign"}}},
                ],
            },
            headers=self.headers,
        )

        resp = self.app.get("/buckets/alice/collections/source", headers=self.headers)
        assert resp.json["data"]["status"] == "signed"
        resp = self.app.get("/buckets/alice/collections/from", headers=self.headers)
        assert resp.json["data"]["status"] == "signed"


class SigningErrorTest(BaseWebTest, unittest.TestCase):
    def test_returns_5xx_if_autograph_cannot_be_reached(self):
        collection_uri = "/buckets/alice/collections/source"
        self.app.app.registry.signers[collection_uri].server_url = "http://0.0.0.0:1234"

        self.app.put_json("/buckets/alice", headers=self.headers)

        resp = self.app.get("/", headers=self.headers)
        self.userid = resp.json["user"]["id"]
        self.app.put_json(
            "/buckets/alice/groups/reviewers",
            {"data": {"members": [self.userid]}},
            headers=self.headers,
        )

        self.app.put_json(
            collection_uri,
            {"data": {"status": "to-sign"}},
            headers=self.headers,
            status=500,
        )


class RecordChangedTest(BaseWebTest, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.headers = get_user_headers("me")

        patch = mock.patch("kinto_remote_settings.signer.backends.autograph.requests")
        self.mock = patch.start()
        self.addCleanup(patch.stop)

        self.collection_uri = "/buckets/alice/collections/source"
        self.records_uri = self.collection_uri + "/records"
        self.app.put_json("/buckets/alice", headers=self.headers)
        self.app.put_json(self.collection_uri, headers=self.headers)

    def test_returns_400_if_contains_float(self):
        self.headers = get_user_headers("me")
        parameters = [
            ({"a": 3.14}, "a"),
            ({"a": {"b": 41.0}}, "a.b"),
        ]
        for data, path in parameters:
            body = {"data": data}
            resp = self.app.post_json(
                self.records_uri, body, headers=self.headers, status=400
            )
            assert path in resp.json["message"]


class SourceCollectionDeletion(BaseWebTest, unittest.TestCase):
    @classmethod
    def get_app_settings(cls, extras=None):
        settings = super(cls, SourceCollectionDeletion).get_app_settings(extras)
        settings["signer.to_review_enabled"] = "true"
        settings["signer.stage.editors_group"] = "something"
        return settings

    def setUp(self):
        super().setUp()

        # Patch calls to Autograph.
        patch = mock.patch("kinto_remote_settings.signer.backends.autograph.requests")
        mocked = patch.start()
        self.addCleanup(patch.stop)
        mocked.post.return_value.json.side_effect = lambda: [
            {
                "signature": uuid.uuid4().hex,
                "hash_algorithm": "",
                "signature_encoding": "",
                "content-signature": "",
                "x5u": "",
                "ref": "",
            }
        ]

        self.headers = get_user_headers("me")
        resp = self.app.get("/", headers=self.headers)
        self.userid = resp.json["user"]["id"]

        self.other_headers = get_user_headers("Sam:Wan Heilss")
        resp = self.app.get("/", headers=self.other_headers)
        self.other_userid = resp.json["user"]["id"]

        self.app.put_json("/buckets/stage", headers=self.headers)

        self.app.put_json(
            "/buckets/stage/groups/something",
            {"data": {"members": [self.other_userid]}},
            headers=self.headers,
        )
        self.app.put_json(
            "/buckets/stage/groups/reviewers",
            {"data": {"members": [self.userid]}},
            headers=self.headers,
        )

        self.create_records_and_sign()

        resp = self.app.get("/buckets/prod/collections/a", headers=self.headers)
        signature_before = resp.json["data"]["signature"]["signature"]
        # Delete source collection.
        self.app.delete("/buckets/stage/collections/a", headers=self.headers)
        # Signature was refreshed.
        resp = self.app.get("/buckets/prod/collections/a", headers=self.headers)
        assert signature_before != resp.json["data"]["signature"]["signature"]

    def create_records_and_sign(self):
        body = {"permissions": {"write": [self.other_userid]}}
        self.app.put_json("/buckets/stage/collections/a", body, headers=self.headers)
        for i in range(5):
            self.app.post_json(
                "/buckets/stage/collections/a/records", headers=self.headers
            )
        self.app.patch_json(
            "/buckets/stage/collections/a",
            {"data": {"status": "to-review"}},
            headers=self.other_headers,
        )
        self.app.patch_json(
            "/buckets/stage/collections/a",
            {"data": {"status": "to-sign"}},
            headers=self.headers,
        )

    def test_unsigned_collection_are_not_affected(self):
        self.app.put_json("/buckets/bid", headers=self.headers)
        self.app.put_json("/buckets/bid/collections/a", headers=self.headers)

        patch = mock.patch(
            "kinto_remote_settings.signer.updater.LocalUpdater.sign_and_update_destination"  # noqa: 501
        )
        mocked = patch.start()
        self.addCleanup(patch.stop)

        self.app.delete("/buckets/bid/collections/a", headers=self.headers)

        assert not mocked.called

    def test_destination_content_is_deleted_when_source_is_deleted(self):
        # Destination is empty.
        resp = self.app.get("/buckets/prod/collections/a/records", headers=self.headers)
        assert len(resp.json["data"]) == 0
        # Tombstones were created.
        resp = self.app.get(
            "/buckets/prod/collections/a/records?_since=0", headers=self.headers
        )
        assert len(resp.json["data"]) == 5
        # Recreate source collection.
        self.create_records_and_sign()
        # Destination has 5 records.
        resp = self.app.get("/buckets/prod/collections/a/records", headers=self.headers)
        assert len(resp.json["data"]) == 5

    def test_preview_content_is_deleted_when_source_is_deleted(self):
        # Preview is empty.
        resp = self.app.get(
            "/buckets/preview/collections/a/records", headers=self.headers
        )
        assert len(resp.json["data"]) == 0
        # Tombstones were created.
        resp = self.app.get(
            "/buckets/preview/collections/a/records?_since=0", headers=self.headers
        )
        assert len(resp.json["data"]) == 5
        # Recreate source collection.
        self.create_records_and_sign()
        # Preview has 5 records.
        resp = self.app.get(
            "/buckets/preview/collections/a/records", headers=self.headers
        )
        assert len(resp.json["data"]) == 5



# TODO: This hardcoded version is a temporary fix to provide a version proprety
# as server metadata. This value will eventually be replaced by a version
# provided when we install `kinto_remote_settings` as a package.
__version__ = "27.0.3"


def includeme(config):
    config.include("kinto_remote_settings.changes")
    config.include("kinto_remote_settings.signer")
